#! /usr/bin/env python3
#
# Copyright (C) 2018  Fx Bricks Inc.
# This file is part of the pfxbrick python module.
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# PFx Brick configuration data helpers

import pfxbrick.pfxdict as pd
from pfxbrick import *
from pfxbrick.pfxhelpers import *


class PFxSettings:
    """
    General settings container class. A member of PFxConfig

    This class contains miscellaneous user preference settings such as
    power saving modes.

    Attributes:
        statusLED (:obj:`int`): status LED mode

        volumeBeep (:obj:`int`): volume beep mode

        autoPowerDown (:obj:`int`): auto power down mode

        lockoutMode (:obj:`int`): IR lockout activation mode

        irAutoOff (:obj:`int`): auto IR disable mode

        bleAutoOff (:obj:`int`): auto BLE disable mode

        bleMotorWhenDisconnect (:obj:`int`): behaviour of motors on BLE disconnect

        bleAdvertPower (:obj:`int`): BLE RF power during advertising

        bleSessionPower (:obj:`int`): BLE RF power during connected session

        notchCount (:obj:`int`): number of motor index speed notches

        notchBounds ([:obj:`int`]): list of monotonic increasing speed notch boundaries
    """

    def __init__(self):
        self.statusLED = 0
        self.volumeBeep = 0
        self.autoPowerDown = 0
        self.lockoutMode = 0
        self.irAutoOff = 0
        self.bleAutoOff = 0
        self.bleMotorWhenDisconnect = 0
        self.bleAdvertPower = 0
        self.bleSessionPower = 0
        self.notchCount = 0
        self.notchBounds = [0, 0, 0, 0, 0, 0, 0]
        self.rapidAccelThr = 0
        self.rapidDecelThr = 0
        self.brakeDecelThr = 0
        self.brakeSpeedThr = 0

    def __eq__(self, other):
        for k, v in self.__dict__.items():
            if k not in other.__dict__:
                return False
            if not v == other.__dict__[k]:
                return False
        return True

    def __str__(self):
        sb = []
        sb.append("Status LED            : %s" % (pd.status_led_dict[self.statusLED]))
        sb.append("Volume Beep           : %s" % (pd.volume_beep_dict[self.volumeBeep]))
        sb.append(
            "Auto Power Down       : %s" % (pd.power_save_dict[self.autoPowerDown])
        )
        sb.append("IR Lockout Mode       : %s" % (pd.lockout_dict[self.lockoutMode]))
        sb.append("IR Auto Off           : %s" % (pd.ir_off_dict[self.irAutoOff]))
        sb.append("BLE Auto Off          : %s" % (pd.ble_off_dict[self.bleAutoOff]))
        sb.append(
            "BLE Motor Disconnect  : %s"
            % (pd.ble_motor_dict[self.bleMotorWhenDisconnect])
        )
        sb.append("BLE Advert Power      : %s" % (self.bleAdvertPower))
        sb.append("BLE Session Power     : %s" % (self.bleSessionPower))
        sb.append("Motor sound notches   : %s" % (self.notchCount))
        mb = "".join("0x{:02X} ".format(x) for x in self.notchBounds)
        sb.append("Motor sound bounds    : %s" % (mb))
        sb.append("Rapid Accel Thr       : %d" % (self.rapidAccelThr))
        sb.append("Rapid Decel Thr       : %d" % (self.rapidDecelThr))
        sb.append("Brake Decel Thr       : %d" % (self.brakeDecelThr))
        sb.append("Brake Speed Thr       : %d" % (self.brakeSpeedThr))
        s = "\n".join(sb)
        return s


class PFxMotor:
    """
    Motor settings container class.

    This class contains motor configuration data for one motor channel.

    Attributes:
        invert (:obj:`boolean`): invert the definition of forward/reverse

        torqueComp (:obj:`boolean`): activate low speed torque compensation with low frequency PWM

        tlgMode (:obj:`boolean`): enable LEGO® Power Functions compatible PWM mode

        accel (:obj:`int`): acceleration factor (0 - 15 max)

        decel (:obj:`int`): deceleration factor (0 - 15 max)

        vmin (:obj:`int`): speed curve minimum mapped speed (0 -> vmid-1)

        vmid (:obj:`int`): speed curve midpoint speed (vmin+1 -> vmax-1)

        vmax (:obj:`int`): speed curve maximum mapped speed (vmid+1 -> 255)
    """

    def __init__(self):
        self.invert = False
        self.torqueComp = False
        self.tlgMode = False
        self.accel = 0
        self.decel = 0
        self.vmin = 0
        self.vmid = 128
        self.vmax = 255

    def __repr__(self):
        s = "invert=%02X torqueComp=%02X tlgMode=%02X" % (
            self.invert,
            self.torqueComp,
            self.tlgMode,
        )
        s = s + "vmin=%02X vmid=%02X vmax=%02X" % (self.vmin, self.vmid, self.vmax)
        s = "%s(%s)" % ("PFxMotor." + self.__class__.__name__, s)
        return s

    def __str__(self):
        sb = []
        sb.append(
            "  Invert : %s  Torque comp : %s  PF mode : %s"
            % (self.invert, self.torqueComp, self.tlgMode)
        )
        sb.append("  Accel  : %s  Decel : %s" % (self.accel, self.decel))
        sb.append(
            "  vMin   : %s  vMid  : %s  vMax : %s" % (self.vmin, self.vmid, self.vmax)
        )
        s = "\n".join(sb)
        return s

    def __eq__(self, other):
        for k, v in self.__dict__.items():
            if k not in other.__dict__:
                return False
            if not v == other.__dict__[k]:
                return False
        return True

    def from_config_byte(self, byte):
        self.invert = set_with_bit(byte, PFX_CFG_MOTOR_INVERT)
        self.torqueComp = set_with_bit(byte, PFX_CFG_MOTOR_TRQCOMP)
        self.tlgMode = set_with_bit(byte, PFX_CFG_MOTOR_TLGMODE)

    def from_speed_bytes(self, msg):
        self.vmin = int(msg[0])
        self.vmid = int(msg[1])
        self.vmax = int(msg[2])
        self.accel = int(msg[3])
        self.decel = int(msg[4])

    def to_config_byte(self):
        v = 0
        if self.invert:
            v |= PFX_CFG_MOTOR_INVERT
        if self.torqueComp:
            v |= PFX_CFG_MOTOR_TRQCOMP
        if self.tlgMode:
            v |= PFX_CFG_MOTOR_TLGMODE
        return v

    def to_speed_bytes(self):
        v = []
        v.append(self.vmin)
        v.append(self.vmid)
        v.append(self.vmax)
        v.append(self.accel)
        v.append(self.decel)
        return v


class PFxLights:
    """
    Light settings container class.

    This class contains default startup brightness data for every light channel.
    All brightness values range from 0 (minimum) to 255 (maximum).

    Attributes:
        defaultBrightness (:obj:`int`): default global brightness, if 0, then individual brightness is used

        startupBrightness ([:obj:`int`]): list of 8 individual startup brightness values for each light output

        pfBrightnessA (:obj:`int`): startup brightness of PF channel A (when used for lights)

        pfBrightnessB (:obj:`int`): startup brightness of PF channel B

        pfBrightnessC (:obj:`int`): startup brightness of PF channel C

        pfBrightnessD (:obj:`int`): startup brightness of PF channel D
    """

    def __init__(self):
        self.defaultBrightness = 0
        self.startupBrightness = [0, 0, 0, 0, 0, 0, 0, 0]
        self.pfBrightnessA = 0
        self.pfBrightnessB = 0
        self.pfBrightnessC = 0
        self.pfBrightnessD = 0

    def __repr__(self):
        sb = "".join("{:02X} ".format(x) for x in self.startupBrightness)
        sp = "".join(
            "{:02X} ".format(x)
            for x in [
                self.pfBrightnessA,
                self.pfBrightnessB,
                self.pfBrightnessC,
                self.pfBrightnessD,
            ]
        )
        s = "defaultBrightness=%02X startupBrightness=%s pfBrightness=%s" % (
            self.defaultBrightness,
            sb,
            sp,
        )
        s = "%s(%s)" % ("PFxLights." + self.__class__.__name__, s)
        return s

    def __eq__(self, other):
        for k, v in self.__dict__.items():
            if k not in other.__dict__:
                return False
            if not v == other.__dict__[k]:
                return False
        return True

    def __str__(self):
        sb = []
        sb.append("Default brightness    : 0x%02X" % (self.defaultBrightness))
        sb.append(
            "Startup brightness    : "
            + "".join("0x{:02X} ".format(x) for x in self.startupBrightness)
        )
        sb.append(
            "PF output brightness  : "
            + "".join(
                "0x{:02X} ".format(x)
                for x in [
                    self.pfBrightnessA,
                    self.pfBrightnessB,
                    self.pfBrightnessC,
                    self.pfBrightnessD,
                ]
            )
        )
        s = "\n".join(sb)
        return s


class PFxAudio:
    """
    Audio settings container class.

    This class contains audio configuration data such as default volume,
    bass, treble, etc.

    Attributes:
        audioDRC (:obj:`boolean`): auto Dynamic Range Control (True/False)

        bass (:obj:`int`): startup bass EQ (-20 to 20 dB)

        treble (:obj:`int`): startup treble EQ (-20 to 20 dB)

        defaultVolume (:obj:`int`): startup volume (0 min - 255 max)
    """

    def __init__(self):
        self.audioDRC = False
        self.bass = 0
        self.treble = 0
        self.defaultVolume = 0

    def __eq__(self, other):
        for k, v in self.__dict__.items():
            if k not in other.__dict__:
                return False
            if not v == other.__dict__[k]:
                return False
        return True

    def __repr__(self):
        s = "drc=%d bass=%d treble=%d" % (self.audioDRC, self.bass, self.treble)
        s = "%s(%s)" % ("PFxAudio." + self.__class__.__name__, s)
        return s

    def __str__(self):
        ds = ""
        if self.audioDRC:
            ds = "ON"
        else:
            ds = "OFF"
        s = "Default Vol: 0x%02X Audio DRC: %s  Bass: 0x%02X  Treble: 0x%02X" % (
            self.defaultVolume,
            ds,
            self.bass,
            self.treble,
        )
        return s


class PFxConfig:
    """
    Top level configuration data container class.

    This class contains catergorized container classes for groups of related settings.
    To change a configuration setting, simply access the setting value using
    a dotted path type notation, e.g. config.lights.startupBrightness[2] = 100

    Attributes:
        settings (:obj:`PFxSettings`): container for general settings.

        motors ([:obj:`PFxMotor`]): list of 4 containers for motor settings

        lights (:obj:`PFxLights`): container for default brightness settings

        audio (:obj:`PFxAudio`): container for audio related settings
    """

    def __init__(self, icd_rev=None):
        self.settings = PFxSettings()
        self.motors = [PFxMotor(), PFxMotor(), PFxMotor(), PFxMotor()]
        self.lights = PFxLights()
        self.audio = PFxAudio()
        if icd_rev is not None:
            self.icd_rev = icd_rev
        else:
            self.icd_rev = "3.38"

    def __eq__(self, other):
        for k, v in self.__dict__.items():
            if k not in other.__dict__:
                return False
            if not v == other.__dict__[k]:
                return False
        return True

    def settings_byte(self):
        v = 0
        if self.settings.statusLED == PFX_CFG_STATLED_OFF:
            v |= PFX_CFG_STATLED_OFF
        else:
            v |= PFX_CFG_STATLED_ON
        if self.settings.volumeBeep == PFX_CFG_VOLBEEP_ON:
            v |= PFX_CFG_VOLBEEP_ON
        else:
            v |= PFX_CFG_VOLBEEP_OFF
        v |= self.settings.autoPowerDown
        v |= self.settings.lockoutMode
        if self.audio.audioDRC == PFX_CFG_AUDIO_DRC_ON:
            v |= PFX_CFG_AUDIO_DRC_ON
        else:
            v |= PFX_CFG_AUDIO_DRC_OFF
        return v

    def from_bytes(self, msg):
        """
        Converts the message string bytes read from the PFx Brick into
        the corresponding data members of this class.
        """
        self.lights.startupBrightness[0] = msg[1]
        self.lights.startupBrightness[1] = msg[2]
        self.lights.startupBrightness[2] = msg[3]
        self.lights.startupBrightness[3] = msg[4]
        self.lights.startupBrightness[4] = msg[5]
        self.lights.startupBrightness[5] = msg[6]
        self.settings.notchCount = msg[7]
        self.settings.notchBounds[0] = msg[8]
        self.settings.notchBounds[1] = msg[9]
        self.settings.notchBounds[2] = msg[10]
        self.settings.notchBounds[3] = msg[11]
        self.settings.notchBounds[4] = msg[12]
        self.settings.notchBounds[5] = msg[13]
        self.settings.notchBounds[6] = msg[14]
        if not is_version_less_than(self.icd_rev, "3.38"):
            self.settings.rapidAccelThr = msg[15]
            self.settings.rapidDecelThr = msg[16]
            self.settings.brakeDecelThr = msg[17]
            self.settings.brakeSpeedThr = msg[18]
        self.settings.irAutoOff = msg[26]
        self.settings.bleAutoOff = msg[27]
        self.settings.bleMotorWhenDisconnect = msg[28]
        self.settings.bleAdvertPower = msg[29]
        self.settings.bleSessionPower = msg[30]
        self.lights.startupBrightness[6] = msg[31]
        self.lights.startupBrightness[7] = msg[32]
        self.lights.pfBrightnessA = msg[33]
        self.lights.pfBrightnessB = msg[34]
        self.audio.bass = msg[35]
        self.audio.treble = msg[36]
        self.settings.statusLED = int(msg[37] & PFX_CFG_STATLED_MASK)
        self.settings.volumeBeep = int(msg[37] & PFX_CFG_VOLBEEP_MASK)
        self.settings.autoPowerDown = int(msg[37] & PFX_CFG_POWERSAVE_MASK)
        self.settings.lockoutMode = int(msg[37] & PFX_CFG_LOCK_MODE_MASK)
        self.audio.audioDRC = int(msg[37] & PFX_CFG_AUDIO_DRC_MASK)
        self.motors[0].from_config_byte(msg[38])
        self.motors[0].from_speed_bytes(msg[39:44])
        self.motors[1].from_config_byte(msg[44])
        self.motors[1].from_speed_bytes(msg[45:50])
        self.motors[2].from_config_byte(msg[50])
        self.motors[2].from_speed_bytes(msg[51:56])
        self.motors[3].from_config_byte(msg[56])
        self.motors[3].from_speed_bytes(msg[57:62])
        self.audio.defaultVolume = msg[62]
        self.lights.defaultBrightness = msg[63]

    def to_bytes(self):
        """
        Converts the data members of this class to the message
        string bytes which can be sent to the PFx Brick.
        """
        msg = []
        msg.append(self.settings.notchCount)
        msg.append(self.settings.notchBounds[0])
        msg.append(self.settings.notchBounds[1])
        msg.append(self.settings.notchBounds[2])
        msg.append(self.settings.notchBounds[3])
        msg.append(self.settings.notchBounds[4])
        msg.append(self.settings.notchBounds[5])
        msg.append(self.settings.notchBounds[6])
        if is_version_less_than(self.icd_rev, "3.38"):
            msg.extend([0] * 11)
        else:
            msg.append(self.settings.rapidAccelThr)
            msg.append(self.settings.rapidDecelThr)
            msg.append(self.settings.brakeDecelThr)
            msg.append(self.settings.brakeSpeedThr)
            msg.extend([0] * 5)
            # include new ICD revision to help the PFx Brick
            # determine what version we are
            major, minor = ver_to_bytes(ICD_REV)
            msg.append(major)
            msg.append(minor)

        msg.append(self.settings.irAutoOff)
        msg.append(self.settings.bleAutoOff)
        msg.append(self.settings.bleMotorWhenDisconnect)
        msg.append(self.settings.bleAdvertPower)
        msg.append(self.settings.bleSessionPower)
        msg.append(self.audio.bass)
        msg.append(self.audio.treble)
        msg.append(self.settings_byte())
        msg.append(self.motors[0].to_config_byte())
        msg.extend(self.motors[0].to_speed_bytes())
        msg.append(self.motors[1].to_config_byte())
        msg.extend(self.motors[1].to_speed_bytes())
        msg.append(self.motors[2].to_config_byte())
        msg.extend(self.motors[2].to_speed_bytes())
        msg.append(self.motors[3].to_config_byte())
        msg.extend(self.motors[3].to_speed_bytes())
        msg.append(self.audio.defaultVolume)
        msg.append(self.lights.defaultBrightness)
        msg.append(self.lights.startupBrightness[0])
        msg.append(self.lights.startupBrightness[1])
        msg.append(self.lights.startupBrightness[2])
        msg.append(self.lights.startupBrightness[3])
        msg.append(self.lights.startupBrightness[4])
        msg.append(self.lights.startupBrightness[5])
        msg.append(self.lights.startupBrightness[6])
        msg.append(self.lights.startupBrightness[7])
        msg.append(self.lights.pfBrightnessA)
        msg.append(self.lights.pfBrightnessB)
        return msg

    def __str__(self):
        sb = []
        sb.append(str(self.settings))
        sb.append(str(self.lights))
        sb.append(str(self.audio))
        for i, motor in enumerate(self.motors):
            sb.append("Motor Channel %d" % (i))
            sb.append(str(motor))
        s = "\n".join(sb)
        return s



#! /usr/bin/env python3
"""
pfxcat - print the contens of a file on the PFx Brick
"""
import argparse
import tempfile
from sys import argv

from pfxbrick import *

TMP_FILE = tempfile.gettempdir() + os.sep + "pfxdump.dat"


def main():
    parser = argparse.ArgumentParser(
        description="PFx Brick print file contents",
        prefix_chars="-+",
    )
    parser.add_argument(
        "file", metavar="file", type=str, help="file name or file ID to show contents"
    )
    parser.add_argument(
        "-s",
        "--serialno",
        default=None,
        help="Specify PFx Brick with serial number (if more than one connected)",
    )
    args = parser.parse_args()
    argsd = vars(args)

    b = get_one_pfxbrick(argsd["serialno"])
    r = b.open()
    if not r:
        exit()
    b.refresh_file_dir()
    f = argsd["file"]
    if f.isnumeric():
        fid = int(f)
    else:
        fid = b.file_id_from_str_or_int(f)
    as_bytes = True if len(argv) == 3 else False
    fd = b.filedir.get_file_dir_entry(fid)
    if fd is not None:
        fs_copy_file_from(
            b, fd, TMP_FILE, show_progress=False, as_bytes=as_bytes, to_console=True
        )
        os.remove(TMP_FILE)
    else:
        print("File %s not found" % (argv[1]))
    b.close()


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
"""
pfxrm - remove a file on the PFx Brick
"""
import argparse

from pfxbrick import *


def main():
    parser = argparse.ArgumentParser(
        description="remove a file from the PFx Brick",
        prefix_chars="-+",
    )
    parser.add_argument(
        "file", metavar="file", type=str, help="file name or file ID to remove"
    )
    parser.add_argument(
        "-s",
        "--serialno",
        default=None,
        help="Specify PFx Brick with serial number (if more than one connected)",
    )
    args = parser.parse_args()
    argsd = vars(args)

    b = get_one_pfxbrick(argsd["serialno"])
    r = b.open()
    if not r:
        exit()
    b.refresh_file_dir()
    f = str(argsd["file"])
    if f.isnumeric():
        f = int(argsd["file"])
    fid = b.file_id_from_str_or_int(f)
    if fid < 0xFF:
        b.remove_file(fid)
    else:
        print("File %s is not found on the PFx Brick" % (argsd["file"]))
    b.close()


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
"""
pfxrestart - restarts the PFx Brick
"""
import argparse

from pfxbrick import *


def main():
    parser = argparse.ArgumentParser(
        description="Restarts the PFx Brick",
        prefix_chars="-+",
    )
    parser.add_argument(
        "-s",
        "--serialno",
        default=None,
        help="Specify PFx Brick with serial number (if more than one connected)",
    )
    parser.add_argument(
        "-x",
        "--halt",
        action="store_true",
        default=False,
        help="Halt all activity on PFx Brick without restarting",
    )
    args = parser.parse_args()
    argsd = vars(args)

    b = get_one_pfxbrick(argsd["serialno"])
    r = b.open()
    if not r:
        exit()
    if argsd["halt"]:
        b.test_action(PFxAction().all_off())
        b.close()
    else:
        b.send_raw_icd_command(
            [
                PFX_USB_CMD_REBOOT,
                PFX_REBOOT_BYTE0,
                PFX_REBOOT_BYTE1,
                PFX_REBOOT_BYTE2,
                PFX_REBOOT_BYTE3,
                PFX_REBOOT_BYTE4,
                PFX_REBOOT_BYTE5,
                PFX_REBOOT_BYTE6,
            ]
        )
        print("PFx Brick restarted")


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
"""
pfxdir - shows the PFx Brick file directory listing
"""
import argparse

from rich import print as colourprint

from pfxbrick import *
from pfxbrick.pfxhelpers import get_one_pfxbrick


def main():
    parser = argparse.ArgumentParser(
        description="PFx Brick file directory listing",
        prefix_chars="-+",
    )
    parser.add_argument(
        "-s",
        "--serialno",
        default=None,
        help="Specify PFx Brick with serial number (if more than one connected)",
    )
    parser.add_argument(
        "-c",
        "--colour",
        action="store_true",
        default=False,
        help="Print directory listing in colour",
    )
    args = parser.parse_args()
    argsd = vars(args)

    b = get_one_pfxbrick(argsd["serialno"])
    r = b.open()
    if not r:
        exit()
    b.refresh_file_dir()
    if argsd["colour"]:
        colourprint(b.filedir.colour_dir())
    else:
        print(b.filedir)
    b.close()


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
#
# Copyright (C) 2018  Fx Bricks Inc.
# This file is part of the pfxbrick python module.
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# PFx Brick dictionary lookup helpers

from pfxbrick.pfx import *

status_dict = {
    PFX_STATUS_NORMAL: "Normal",
    PFX_STATUS_NORMAL_PENDING: "Normal, upgrade pending",
    PFX_STATUS_SERVICE: "Service",
    PFX_STATUS_SERVICE_PENDING: "Service, upgrade pending",
    PFX_STATUS_SERVICE_BUSY: "Service, busy",
}

err_dict = {
    PFX_ERR_VERIFY_FAIL: "Verify failed",
    PFX_ERR_TRANSFER_FILE_EXISTS: "File exists",
    PFX_ERR_TRANSFER_TOO_BIG: "File too big",
    PFX_ERR_TRANSFER_INVALID: "Transfer request invalid",
    PFX_ERR_TRANSFER_FILE_NOT_FOUND: "File not found",
    PFX_ERR_TRANSFER_ERROR: "Transfer error",
    PFX_ERR_TRANSFER_CRC_MISMATCH: "CRC mismatch",
    PFX_ERR_TRANSFER_BUSY_WAIT: "Busy, try again",
    PFX_ERR_TRANSFER_LUT_FULL: "LUT full",
    PFX_ERR_TRANSFER_COMPLETE: "Transfer complete",
    PFX_ERR_UPGRADE_FAIL: "Upgrade failed",
    PFX_ERR_FILE_SYSTEM_ERR: "File system error",
    PFX_ERR_FILE_INVALID: "File is invalid",
    PFX_ERR_FILE_OUT_OF_RANGE: "File access is out of range",
    PFX_ERR_FILE_READ_ONLY: "File is read only",
    PFX_ERR_FILE_TOO_BIG: "File is too big",
    PFX_ERR_FILE_NOT_FOUND: "File not found",
    PFX_ERR_FILE_NOT_UNIQUE: "File ID is not unique",
    PFX_ERR_FILE_LOCKED_BUSY: "File system is locked or busy",
    PFX_ERR_FILE_SYSTEM_FULL: "File system is full",
    PFX_ERR_FILE_SYSTEM_TIMEOUT: "File system operation timed out",
    PFX_ERR_FILE_INVALID_ADDRESS: "File system accessed invalid address",
    PFX_ERR_FILE_NEXT_SECTOR: "File system error reading next sector",
    PFX_ERR_FILE_ACCESS_DENIED: "File system access denied",
    PFX_ERR_BLE_FAULT: "Bluetooth fault",
    PFX_ERR_SPKR_SHORTCIR_FAULT: "Short circuit on speaker output",
    PFX_ERR_DAC_OVERTEMP_FAULT: "Overtemperature reported by DAC",
    PFX_ERR_TRAP_BROWNOUT_RST: "Brown-out reset trap",
    PFX_ERR_TRAP_CONFLICT: "Reset trap conflict",
    PFX_ERR_TRAP_ILLEGAL_OPCODE: "Reset trap due to illegal OP code",
    PFX_ERR_TRAP_CONFIG_MISMATCH: "Reset trap due to configuration mismatch",
}


def get_err_str(err):
    if err in err_dict:
        return err_dict[err]
    else:
        return "Unknown error"


status_led_dict = {
    PFX_CFG_STATLED_OFF: "Normally OFF, blink with activity",
    PFX_CFG_STATLED_ON: "Normally ON, blink with activity",
}

volume_beep_dict = {PFX_CFG_VOLBEEP_ON: "ON", PFX_CFG_VOLBEEP_OFF: "OFF"}

power_save_dict = {
    PFX_CFG_POWERSAVE_OFF: "OFF",
    PFX_CFG_POWERSAVE_30M: "30 min",
    PFX_CFG_POWERSAVE_60M: "60 min",
    PFX_CFG_POWERSAVE_3HR: "3 hrs",
}

lockout_dict = {
    PFX_CFG_LOCKOUT_INH: "Inhibit",
    PFX_CFG_LOCKOUT_CH1: "Toggle from ch 1",
    PFX_CFG_LOCKOUT_ALLCH: "Toggle from any ch",
}

ir_off_dict = {
    PFX_CFG_IRAUTO_OFF_NEVER: "Never",
    PFX_CFG_IRAUTO_OFF_1MIN: "After 1 min",
    PFX_CFG_IRAUTO_OFF_5MIN: "After 5 min",
    PFX_CFG_IRAUTO_OFF_IMMEDIATE: "Immediately",
}

ble_off_dict = {
    PFX_CFG_BLEAUTO_OFF_NEVER: "Never",
    PFX_CFG_BLEAUTO_OFF_1MIN: "After 1 min",
    PFX_CFG_BLEAUTO_OFF_5MIN: "After 5 min",
    PFX_CFG_BLEAUTO_OFF_IMMEDIATE: "Immediately",
}

ble_motor_dict = {
    PFX_CFG_BLE_MOTOR_CONTINUE: "Continue operation",
    PFX_CFG_BLE_MOTOR_STOP: "Stop motors",
}

command_dict = {
    EVT_COMMAND_NONE: "None",
    EVT_COMMAND_ALL_OFF: "All OFF",
    EVT_COMMAND_IR_LOCKOUT_ON: "IR Lockout ON",
    EVT_COMMAND_IR_LOCKOUT_OFF: "IR Lockout OFF",
    EVT_COMMAND_IR_LOCK_TOGGLE: "Toggle IR Lockout",
    EVT_COMMAND_ALL_MOTORS_OFF: "All Motors OFF",
    EVT_COMMAND_ALL_LIGHTS_OFF: "All Lights OFF",
    EVT_COMMAND_ALL_AUDIO_OFF: "All Audio OFF",
    EVT_COMMAND_RESTART: "Restart",
    EVT_COMMAND_RUN_SCRIPT: "Run Script",
}

motor_action_dict = {
    EVT_MOTOR_ESTOP: "Emergency Stop",
    EVT_MOTOR_STOP: "Stop",
    EVT_MOTOR_INC_SPD: "Increase Speed",
    EVT_MOTOR_DEC_SPD: "Decrease Speed",
    EVT_MOTOR_INC_SPD_BI: "Increase Speed (bi-dir)",
    EVT_MOTOR_DEC_SPD_BI: "Decrease Speed (bi-dir)",
    EVT_MOTOR_CHANGE_DIR: "Change Direction",
    EVT_MOTOR_SET_SPD: "Set Speed",
    EVT_MOTOR_SET_SPD_TIMED: "Set Speed with Duration",
    EVT_MOTOR_OSCILLATE: "Oscillate",
    EVT_MOTOR_OSCILLATE_BIDIR: "Oscillate (bi-dir)",
    EVT_MOTOR_OSCILLATE_BIDIR_WAIT: "Oscillate (bi-dir) with wait",
    EVT_MOTOR_RANDOM: "Random",
    EVT_MOTOR_RANDOM_BIDIR: "Random (bi-dir)",
    EVT_MOTOR_SOUND_MODULATED: "Sound Modulated",
    EVT_MOTOR_SET_SERVO: "Set Servo",
}

ind_lightfx_dict = {
    EVT_LIGHTFX_NONE: "None",
    EVT_LIGHTFX_ON_OFF_TOGGLE: "On/Off",
    EVT_LIGHTFX_INC_BRIGHT: "Increase Brightness",
    EVT_LIGHTFX_DEC_BRIGHT: "Decrease Brightness",
    EVT_LIGHTFX_SET_BRIGHT: "Set Brightness",
    EVT_LIGHTFX_FLASH50_P: "Flash 50%, positive",
    EVT_LIGHTFX_FLASH50_N: "Flash 50%, negative",
    EVT_LIGHTFX_STROBE_P: "Strobe Flasher, positive",
    EVT_LIGHTFX_STROBE_N: "Strobe Flasher, negative",
    EVT_LIGHTFX_GYRALITE_P: "Gyralite, postive",
    EVT_LIGHTFX_GYRALITE_N: "Gyralite, negative",
    EVT_LIGHTFX_FLICKER: "Flicker",
    EVT_LIGHTFX_RAND_BLINK: "Random Blink",
    EVT_LIGHTFX_PHOTON_TORP: "Photon Torpedo",
    EVT_LIGHTFX_LASER_PULSE: "Laser Pulse",
    EVT_LIGHTFX_ENGINE_GLOW: "Engine Glow",
    EVT_LIGHTFX_LIGHTHOUSE: "Light House",
    EVT_LIGHTFX_BROKEN_LIGHT: "Broken Light",
    EVT_LIGHTFX_STATUS_IND: "Status Indicator",
    EVT_LIGHTFX_SOUND_MOD: "Sound Modulated",
    EVT_LIGHTFX_MOTOR_MOD: "Motor Modulated",
}

combo_lightfx_dict = {
    EVT_COMBOFX_NONE: "None",
    EVT_COMBOFX_LIN_SWEEP: "Linear Sweep",
    EVT_COMBOFX_BARGRAPH: "Bargraph",
    EVT_COMBOFX_KNIGHTRIDER: "Knight Rider",
    EVT_COMBOFX_EMCY_TWSONIC: "Twinsonic Style Flashers",
    EVT_COMBOFX_EMCY_WHELEN: "Strobe Style Flashers",
    EVT_COMBOFX_TIMES_SQ: "Times Square",
    EVT_COMBOFX_NOISE: "Noise",
    EVT_COMBOFX_TWINKLE_STAR: "Twinkling Stars",
    EVT_COMBOFX_TRAFFIC_SIG: "Traffic Lights",
    EVT_COMBOFX_SOUND_BAR: "Sound Bar",
    EVT_COMBOFX_ALT_FLASH: "Alternating Flashers",
    EVT_COMBOFX_LAVA_LAMP: "Lava Lamp",
    EVT_COMBOFX_LASER_CANNON: "Laser Cannon",
    EVT_COMBOFX_DRAGSTER: "Dragster Starter",
    EVT_COMBOFX_RUNWAY: "Airport Runway",
    EVT_COMBOFX_FORMULA1: "Formula 1 Lights",
}

soundfx_dict = {
    EVT_SOUND_NONE: "None",
    EVT_SOUND_INC_VOL: "Increase Volume",
    EVT_SOUND_DEC_VOL: "Decrease Volume",
    EVT_SOUND_SET_VOL: "Set Volume",
    EVT_SOUND_PLAY_ONCE: "Play Once",
    EVT_SOUND_PLAY_CONT: "Play Repeat",
    EVT_SOUND_PLAY_NTIMES: "Play N Times",
    EVT_SOUND_PLAY_DUR: "Play for Duration",
    EVT_SOUND_PLAY_PITCH: "Speed Modulated Pitch",
    EVT_SOUND_PLAY_GATED: "Speed Modulated Gated",
    EVT_SOUND_PLAY_AM: "Speed Modulated Volume",
    EVT_SOUND_STOP: "Stop Play",
    EVT_SOUND_PLAY_IDX_MOTOR: "Speed Indexed",
    EVT_SOUND_PLAY_RAND: "Random Play",
    EVT_SOUND_PLAY_PAUSE_FLAG: "Play/Pause",
}

evtid_dict = {
    EVT_ID_8879_TWO_BUTTONS: "Left + Right Button",
    EVT_ID_8879_LEFT_BUTTON: "Left Button",
    EVT_ID_8879_RIGHT_BUTTON: "Right Button",
    EVT_ID_8879_LEFT_INC: "Left Wheel Inc",
    EVT_ID_8879_LEFT_DEC: "Left Wheel Dec",
    EVT_ID_8879_RIGHT_INC: "Right Wheel Inc",
    EVT_ID_8879_RIGHT_DEC: "Right Wheel Dec",
    EVT_ID_8885_LEFT_FWD: "Left Joystick Fwd",
    EVT_ID_8885_LEFT_REV: "Left Joystick Rev",
    EVT_ID_8885_RIGHT_FWD: "Right Joystick Fwd",
    EVT_ID_8885_RIGHT_REV: "Right Joystick Rev",
    EVT_ID_8885_LEFT_CTROFF: "Left Joystick Centre",
    EVT_ID_8885_RIGHT_CTROFF: "Right Joystick Centre",
    EVT_ID_EV3_BEACON: "EV3 Beacon",
    EVT_ID_TEST_EVENT: "Test Event",
    EVT_ID_STARTUP_EVENT: "Startup Events 1-4",
    EVT_ID_STARTUP_EVENT2: "Startup Events 5-8",
    EVT_ID_RCTRAIN_UP: "RC Train Up",
    EVT_ID_RCTRAIN_DOWN: "RC Train Down",
    EVT_ID_RCTRAIN_STOP: "RC Train Stop",
    EVT_ID_RCTRAIN_HORN: "RC Train Horn",
    EVT_STARTUP_EVENT1: "Startup Event 1",
    EVT_STARTUP_EVENT2: "Startup Event 2",
    EVT_STARTUP_EVENT3: "Startup Event 3",
    EVT_STARTUP_EVENT4: "Startup Event 4",
    EVT_STARTUP_EVENT5: "Startup Event 5",
    EVT_STARTUP_EVENT6: "Startup Event 6",
    EVT_STARTUP_EVENT7: "Startup Event 7",
    EVT_STARTUP_EVENT8: "Startup Event 8",
    EVT_BUTTON_PRESS: "Button",
    EVT_BUTTON_LONGPRESS: "Long Button",
    EVT_BUTTON_DOWN: "Button Down",
    EVT_BUTTON_UP: "Button Up",
    EVT_BLE_CONNECT: "BLE Connect",
    EVT_BLE_DISCONNECT: "BLE Disconnect",
    EVT_USB_CONNECT: "USB Connect",
    EVT_USB_DISCONNECT: "USB Disconnect",
}

fileid_dict = {
    PFX_SOUND_IDX_CHANGE_DIR: "Dir Change",
    PFX_SOUND_IDX_SET_OFF: "Set Off",
    PFX_SOUND_IDX_RAPID_ACCEL: "Rapid Accel",
    PFX_SOUND_IDX_RAPID_DECEL: "Rapid Decel",
    PFX_SOUND_IDX_BRAKE_STOP: "Brake",
    PFX_SOUND_IDX_STARTUP: "Startup",
    PFX_SOUND_IDX_SHUTDOWN: "Shutdown",
    PFX_SOUND_IDX_NOTCH1_LOOP: "Loop 1",
    PFX_SOUND_IDX_NOTCH2_LOOP: "Loop 2",
    PFX_SOUND_IDX_NOTCH3_LOOP: "Loop 3",
    PFX_SOUND_IDX_NOTCH4_LOOP: "Loop 4",
    PFX_SOUND_IDX_NOTCH5_LOOP: "Loop 5",
    PFX_SOUND_IDX_NOTCH6_LOOP: "Loop 6",
    PFX_SOUND_IDX_NOTCH7_LOOP: "Loop 7",
    PFX_SOUND_IDX_NOTCH8_LOOP: "Loop 8",
    PFX_SOUND_IDX_ACCEL1: "Accel 1-2",
    PFX_SOUND_IDX_ACCEL2: "Accel 2-3",
    PFX_SOUND_IDX_ACCEL3: "Accel 3-4",
    PFX_SOUND_IDX_ACCEL4: "Accel 4-5",
    PFX_SOUND_IDX_ACCEL5: "Accel 5-6",
    PFX_SOUND_IDX_ACCEL6: "Accel 6-7",
    PFX_SOUND_IDX_ACCEL7: "Accel 7-8",
    PFX_SOUND_IDX_DECEL1: "Decel 2-1",
    PFX_SOUND_IDX_DECEL2: "Decel 3-2",
    PFX_SOUND_IDX_DECEL3: "Decel 4-3",
    PFX_SOUND_IDX_DECEL4: "Decel 5-4",
    PFX_SOUND_IDX_DECEL5: "Decel 6-5",
    PFX_SOUND_IDX_DECEL6: "Decel 7-6",
    PFX_SOUND_IDX_DECEL7: "Decel 8-7",
    PFX_SOUND_IDX_GATED_NOTCH11: "Gated Loop 11",
    PFX_SOUND_IDX_GATED_NOTCH12: "Gated Loop 12",
    PFX_SOUND_IDX_GATED_NOTCH13: "Gated Loop 13",
    PFX_SOUND_IDX_GATED_NOTCH14: "Gated Loop 14",
    PFX_SOUND_IDX_GATED_NOTCH21: "Gated Loop 21",
    PFX_SOUND_IDX_GATED_NOTCH22: "Gated Loop 22",
    PFX_SOUND_IDX_GATED_NOTCH23: "Gated Loop 23",
    PFX_SOUND_IDX_GATED_NOTCH24: "Gated Loop 24",
    PFX_SOUND_IDX_GATED_NOTCH31: "Gated Loop 31",
    PFX_SOUND_IDX_GATED_NOTCH32: "Gated Loop 32",
    PFX_SOUND_IDX_GATED_NOTCH33: "Gated Loop 33",
    PFX_SOUND_IDX_GATED_NOTCH34: "Gated Loop 34",
    PFX_SOUND_IDX_GATED_NOTCH41: "Gated Loop 41",
    PFX_SOUND_IDX_GATED_NOTCH42: "Gated Loop 42",
    PFX_SOUND_IDX_GATED_NOTCH43: "Gated Loop 43",
    PFX_SOUND_IDX_GATED_NOTCH44: "Gated Loop 44",
}

file_attr_dict = {
    PFX_SOUND_ATTR_MOTOR_LOOP_N1: "Loop 1",
    PFX_SOUND_ATTR_MOTOR_LOOP_N2: "Loop 2",
    PFX_SOUND_ATTR_MOTOR_LOOP_N3: "Loop 3",
    PFX_SOUND_ATTR_MOTOR_LOOP_N4: "Loop 4",
    PFX_SOUND_ATTR_MOTOR_LOOP_N5: "Loop 5",
    PFX_SOUND_ATTR_MOTOR_LOOP_N6: "Loop 6",
    PFX_SOUND_ATTR_MOTOR_LOOP_N7: "Loop 7",
    PFX_SOUND_ATTR_MOTOR_LOOP_N8: "Loop 8",
    PFX_SOUND_ATTR_MOTOR_ACCEL12: "Accel 1-2",
    PFX_SOUND_ATTR_MOTOR_ACCEL23: "Accel 2-3",
    PFX_SOUND_ATTR_MOTOR_ACCEL34: "Accel 3-4",
    PFX_SOUND_ATTR_MOTOR_ACCEL45: "Accel 4-5",
    PFX_SOUND_ATTR_MOTOR_ACCEL56: "Accel 5-6",
    PFX_SOUND_ATTR_MOTOR_ACCEL67: "Accel 6-7",
    PFX_SOUND_ATTR_MOTOR_ACCEL78: "Accel 7-8",
    PFX_SOUND_ATTR_MOTOR_STARTUP: "Startup",
    PFX_SOUND_ATTR_MOTOR_DECEL21: "Decel 2-1",
    PFX_SOUND_ATTR_MOTOR_DECEL32: "Decel 3-2",
    PFX_SOUND_ATTR_MOTOR_DECEL43: "Decel 4-3",
    PFX_SOUND_ATTR_MOTOR_DECEL54: "Decel 5-4",
    PFX_SOUND_ATTR_MOTOR_DECEL65: "Decel 6-5",
    PFX_SOUND_ATTR_MOTOR_DECEL76: "Decel 7-6",
    PFX_SOUND_ATTR_MOTOR_DECEL87: "Decel 8-7",
    PFX_SOUND_ATTR_MOTOR_SHUTDOWN: "Shutdown",
}



#! /usr/bin/env python3
#
# Copyright (C) 2018  Fx Bricks Inc.
# This file is part of the pfxbrick python module.
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


# PFx Brick python API

import hid
from bleak import BleakClient, BleakScanner

from pfxbrick import *
from pfxbrick.pfxfiles import fs_format
from pfxbrick.pfxhelpers import *
from pfxbrick.pfxmsg import *


def find_bricks(show_list=False):
    """
    Enumerate and optionally print a list PFx Bricks currently connected to the USB bus.

    :param boolean show_list: optionally print a list of enumerated PFx Bricks
    :returns: [:obj:`str`] a list of PFx Brick serial numbers
    """
    numBricks = 0
    serials = []
    for dev in hid.enumerate():
        if (
            dev["vendor_id"] == PFX_USB_VENDOR_ID
            and dev["product_id"] == PFX_USB_PRODUCT_ID
        ):
            if dev["serial_number"] not in serials:
                numBricks += 1
                serials.append(dev["serial_number"])
                h = hid.device()
                h.open(PFX_USB_VENDOR_ID, PFX_USB_PRODUCT_ID, dev["serial_number"])
                usb_prod_str = h.get_product_string()
                usb_serno_str = h.get_serial_number_string()
                if show_list == True:
                    print(
                        "%d. %s, Serial No: %s"
                        % (numBricks, usb_prod_str, usb_serno_str)
                    )
                h.close()
    return serials


class PFxBrick:
    """
    Top level PFx Brick object class.

    This class is used to initialize and maintain a communication session
    with a USB connected PFx Brick. Many convenient methods are provided
    to perform tasks such as changing configuration, accessing the file
    system, initiating actions, and more.

    Attributes:
        product_id (:obj:`str`): product ID code reported by the PFx Brick (e.g. 'A204')

        serial_no (:obj:`str`): serial number reported by the PFx Brick, usually 8 digit hexadecimal

        product_desc (:obj:`str`): product descriptor reported by the PFx Brick

        firmware_ver (:obj:`str`): firmware version number reported, 4-digit hex BCD, e.g. '0134' represents v.1.34

        firmware_build (:obj:`str`): firmware build number reported, 4-digit hex BCD

        icd_rev (:obj:`str`): ICD revision number reported, 4-digit hex BCD, e.g. '0336' represents v.3.36
        status (:obj:`int`): status code reported, distinguishes normal and service/recovery mode

        error (:obj:`int`): error code reported

        usb_vid (:obj:`int`): fixed to PFX_USB_VENDOR_ID representing the official USB VID assigned to the PFx Brick

        usb_pid (:obj:`int`): fixed to PFX_USB_PRODUCT_ID representing the official USB PID assigned to the PFx Brick

        usb_manu_str (:obj:`str`): the manufacturer string reported to the host USB interface

        usb_prod_str (:obj:`str`): the product descriptor string reported to the host USB interface

        usb_serno_str (:obj:`str`): the product serial number string reported to the host USB interface

        dev (:obj:`device`): a device handle to the HIDAPI cdef class device

        is_open (:obj:`boolean`): a flag indicating connected session status

        name (:obj:`str`): user defined name of the PFx Brick

        config (:obj:`PFxConfig`): child class to store configuration and settings

        filedir (:obj:`PFxDir`): child class to store the file system directory
    """

    def __init__(self, serial_no=None):
        self.product_id = ""
        self.serial_no = ""
        self.product_desc = ""
        self.firmware_ver = ""
        self.firmware_build = ""
        self.icd_rev = ""
        self.status = 0
        self.error = 0
        self.usb_vid = PFX_USB_VENDOR_ID
        self.usb_pid = PFX_USB_PRODUCT_ID
        self.usb_manu_str = ""
        self.usb_prod_str = ""
        self.usb_serno_str = ""
        self.dev = None
        self.is_open = False
        self.has_bluetooth = False
        self.name = ""

        self.config = PFxConfig()
        self.filedir = PFxDir()
        self.state = PFxState()

        if serial_no is not None:
            self.open(ser_no=serial_no)

    def open(self, ser_no=None):
        """
        Opens a USB communication session with a PFx Brick. If multiple PFx Bricks are
        connected, then a serial number must be specified to connect to a unique PFx Brick.

        :param ser_no: optional serial number to specify a particular PFx Brick if multiple connected
        :returns: boolean indicating open session result
        """
        if ser_no is not None:
            ser_no = ser_no.upper()
        if not self.is_open:
            numBricks = 0
            serials = []
            for dev in hid.enumerate():
                if (
                    dev["vendor_id"] == PFX_USB_VENDOR_ID
                    and dev["product_id"] == PFX_USB_PRODUCT_ID
                ):
                    if dev["serial_number"] not in serials:
                        numBricks += 1
                        serials.append(dev["serial_number"])
            if ser_no is not None and ser_no not in serials:
                print("The PFx Brick with serial number %s was not found." % (ser_no))
            else:
                if numBricks == 0:
                    print("No PFx Bricks are currently connected.")
                elif numBricks > 1 and ser_no is None:
                    print(
                        "There are multiple PFx Bricks connected. Therefore a serial number is required to specify which PFx Brick to connect to."
                    )
                else:
                    self.dev = hid.device()
                    self.dev.open(PFX_USB_VENDOR_ID, PFX_USB_PRODUCT_ID, ser_no)
                    self.usb_manu_str = self.dev.get_manufacturer_string()
                    self.usb_prod_str = self.dev.get_product_string()
                    self.usb_serno_str = self.dev.get_serial_number_string()
                    self.is_open = True
                    self.get_icd_rev()
                    self.config.icd_rev = self.icd_rev
                    self.get_status()
                    if self.product_id in ["A204", "A208", "A216"]:
                        self.has_bluetooth = True
        return self.is_open

    def close(self):
        """
        Closes a USB communication session with a PFx Brick.
        """
        if self.is_open:
            self.dev.close()

    def get_icd_rev(self, silent=False):
        """
        Requests the version of Interface Control Document (ICD)
        the connected PFx Brick supports using the `PFX_CMD_GET_ICD_REV`
        ICD message.  The resulting version number is stored in
        this class and also returned.

        :param boolean silent: flag to optionally silence the status LED blink

        :returns: :obj:`str` ICD version, e.g. "3.38"
        """
        res = cmd_get_icd_rev(self.dev, silent)
        self.icd_rev = uint16_tover(res[1], res[2])
        return self.icd_rev

    def get_status(self):
        """
        Requests the top level operational status of the PFx Brick

        Summary operating status is obtained from the PFx Brick using
        the `PFX_CMD_GET_STATUS ICD` message.  The resulting
        status data is stored in this class and can be queried
        with class member access methods such as `status_str`.

        :returns: None
        """
        res = cmd_get_status(self.dev)
        if res:
            self.status = res[1]
            self.error = res[2]
            self.product_id = uint16_tostr(res[7], res[8])
            self.serial_no = uint32_tostr(res[9], res[10], res[11], res[12])
            self.product_desc = bytes(res[13:37]).decode("utf-8")
            self.firmware_ver = uint16_tover(res[37], res[38])
            self.firmware_build = uint16_tostr(res[39], res[40])

    def status_str(self):
        """
        Returns a summary status string of the PFx Brick

        A human readable string of PFx Brick basic identification and
        and operating status.  A call to the `get_status` method is recommended
        to ensure that the status information is recent.

        :returns: :obj:`str` status string
        """
        s = []
        s.append("USB vendor ID         : 0x%04X" % (self.usb_vid))
        s.append("USB product ID        : 0x%04X" % (self.usb_pid))
        s.append("USB product desc      : %s" % (self.usb_prod_str))
        s.append("USB manufacturer      : %s" % (self.usb_manu_str))
        s.append(
            "PFx Brick product ID  : %s, %s" % (self.product_id, self.product_desc)
        )
        s.append("Serial number         : %s" % (self.serial_no))
        s.append(
            "Firmware version      : %s build %s"
            % (self.firmware_ver, self.firmware_build)
        )
        s.append(
            "Status                : 0x%02X %s"
            % (self.status, get_status_str(self.status))
        )
        s.append(
            "Errors                : 0x%02X %s"
            % (self.error, get_error_str(self.error))
        )
        return "\n".join(s)

    def print_status(self):
        """
        Prints the top level operational status information retrieved
        by a previous call to the get_status method.
        """
        print(self.status_str())

    def get_config(self):
        """
        Retrieves configuration settings from the PFx Brick using
        the `PFX_CMD_GET_CONFIG` ICD message. The configuration data
        is stored in the :obj:`PFxBrick.config` class member variable.
        """
        res = cmd_get_config(self.dev)
        if res:
            self.config.from_bytes(res)

    def print_config(self):
        """
        Prints a summary representation of the PFx Brick configuration
        settings which were retrieved by a previous call to get_config.
        """
        print(str(self.config))

    def set_config(self):
        """
        Writes the contents of the :obj:`PFxConfig` data structure class to
        the PFx Brick using the `PFX_CMD_SET_CONFIG ICD` message.

        It is recommended that the configuration be read from the
        PFx Brick (using `get_config`) before any changes are made to
        the configuration and written back. This ensures that any
        configuration settings which are not desired to be changed
        are left in the same state.
        """
        res = cmd_set_config(self.dev, self.config.to_bytes())

    def reset_factory_config(self):
        """
        Resets the PFx Brick configuration settings to factory defaults.
        """
        res = cmd_set_factory_defaults(self.dev)

    def get_name(self):
        """
        Retrieves the user defined name of the PFx Brick using
        the `PFX_CMD_GET_NAME ICD` message. The name is stored in
        the name class variable as a UTF-8 string.

        :returns: :obj:`str` user defined name
        """
        res = cmd_get_name(self.dev)
        if res:
            self.name = safe_unicode_str(res[1:25])
        return self.name

    def set_name(self, name):
        """
        Sets the user defined name of the PFx Brick using the
        `PFX_CMD_SET_NAME ICD` message.

        :param name: :obj:`str` new name to set (up to 24 character bytes, UTF-8)
        """
        res = cmd_set_name(self.dev, name)

    def get_action_by_address(self, address):
        """
        Retrieves a stored action indexed by address rather than a
        combination of eventID and IR channel.  The address is converted into a
        [eventID, IR channel] pair and the get_action method is
        called with this function as a convenient wrapper.

        :param address: :obj:`int` event/action LUT address (0 - 0x7F)
        :returns: :obj:`PFxAction` class filled with retrieved LUT data
        """
        if address > EVT_LUT_MAX:
            print("Requested action at address %02X is out of range" % (address))
            return None
        else:
            evt, ch = address_to_evtch(address)
            return self.get_action(evt, ch)

    def get_action(self, evtID, ch):
        """
        Retrieves the stored action associated with a particular
        [eventID / IR channel] event. The eventID and channel value
        form a composite address pointer into the event/action LUT
        in the PFx Brick. The address to the LUT is formed as:

        Address[5:2] = event ID
        Address[1:0] = channel

        :param evtID: :obj:`int` event ID LUT address component (0 - 0x20)
        :param channel: :obj:`int` channel index LUT address component (0 - 3)
        :returns: :obj:`PFxAction` class filled with retrieved LUT data
        """
        if ch > 3 or evtID > EVT_ID_MAX:
            print("Requested action (id=%02X, ch=%02X) is out of range" % (evtID, ch))
            return None
        else:
            res = cmd_get_event_action(self.dev, evtID, ch)
            action = PFxAction()
            if res:
                action.from_bytes(res)
            return action

    def set_action_by_address(self, address, action):
        """
        Sets a new stored action in the event/action LUT at the
        address specified. The address is converted into a
        [eventID, IR channel] pair and the set_action method is
        called with this function as a convenient wrapper.

        :param address: :obj:`int` event/action LUT address (0 - 0x7F)
        :param action: :obj:`PFxAction` action data structure class
        """
        if address > EVT_LUT_MAX:
            print("Requested action at address %02X is out of range" % (address))
            return None
        else:
            evt, ch = address_to_evtch(address)
            self.set_action(evt, ch, action)

    def set_action(self, evtID, ch, action):
        """
        Sets a new stored action associated with a particular
        [eventID / IR channel] event. The eventID and channel value
        form a composite address pointer into the event/action LUT
        in the PFx Brick. The address to the LUT is formed as:

        Address[5:2] = event ID
        Address[1:0] = channel

        :param evtID: :obj:`int` event ID LUT address component (0 - 0x20)
        :param ch: :obj:`int` channel index LUT address component (0 - 3)
        :param action: :obj:`PFxAction` action data structure class
        """
        if ch > 3 or evtID > EVT_ID_MAX:
            print("Requested action (id=%02X, ch=%02X) is out of range" % (evtID, ch))
            return None
        else:
            res = cmd_set_event_action(self.dev, evtID, ch, action.to_bytes())

    def test_action(self, action):
        """
        Executes a passed action data structure. This function is
        used to "test" actions to see how they behave. The passed
        action is not stored in the event/action LUT.

        :param action: :obj:`PFxAction` action data structure class
        """
        res = cmd_test_action(self.dev, action.to_bytes())

    def clear_action_by_address(self, address):
        """
        Clears a stored action in the event/action LUT at the
        address specified. The address is converted into a
        [eventID, IR channel] pair and the set_action method is
        called with this function as a convenient wrapper.

        :param address: :obj:`int` event/action LUT address (0 - 0x7F)
                        :obj:`list,tuple,range` specify a list or range of addresses
        """
        if isinstance(address, (list, tuple)):
            addresses = address
        elif isinstance(address, range):
            addresses = [x for x in range]
        else:
            addresses = [address]
        for a in addresses:
            if a > EVT_LUT_MAX:
                print("Requested action at address %02X is out of range" % (a))
                return None
            else:
                evt, ch = address_to_evtch(a)
                self.clear_action(evt, ch)

    def clear_action(self, evtID, ch):
        """
        Clears a stored action associated with a particular
        [eventID / IR channel] event. The eventID and channel value
        form a composite address pointer into the event/action LUT
        in the PFx Brick. The address to the LUT is formed as:

        Address[5:2] = event ID
        Address[1:0] = channel

        :param evtID: :obj:`int` event ID LUT address component (0 - 0x20)
        :param ch: :obj:`int` channel index LUT address component (0 - 3)
        """
        if ch > 3 or evtID > EVT_ID_MAX:
            print("Requested action (id=%02X, ch=%02X) is out of range" % (evtID, ch))
            return None
        else:
            # set to an empty PFxAction to clear
            res = cmd_set_event_action(self.dev, evtID, ch, PFxAction().to_bytes())

    def find_startup_action(self, lightfx=None, soundfx=None, motorfx=None):
        """
        Finds a startup action with one or more specified light, motor, or
        sound fx types.

        :param lightfx: :obj:`int` optional lightfx id
        :param soundfx: :obj:`int` optional soundfx id
        :param motorfx: :obj:`int` optional motorfx id

        :returns: :obj:`PFxAction` action data structure class matching the desired type(s)
        """
        req_matches = 0
        if lightfx is not None:
            req_matches += 1
        if soundfx is not None:
            req_matches += 1
        if motorfx is not None:
            req_matches += 1
        for e in range(EVT_STARTUP_EVENT1, EVT_STARTUP_EVENT8 + 1):
            matches = 0
            action = self.get_action_by_address(e)
            if lightfx is not None and action.lightFxId == lightfx:
                matches += 1
            if soundfx is not None and action.soundFxId == soundfx:
                matches += 1
            if motorfx is not None and action.motorActionId == motorfx:
                matches += 1
            if matches == req_matches:
                return action
        return None

    def set_motor_speed(self, ch, speed, duration=None):
        """
        A convenience wrapper for `PFxAction().set_motor_speed`

        :param ch: [:obj:`int`] a list of motor channels (1-4)
        :param speed: :obj:`int` desired motor speed (-100 to +100)
        :param duration: :obj:`float` optional duration (in seconds) to run motor, runs indefinitely if not specified

        If the duration value is specified, it represents the desired motor
        run time in seconds. Note that this value will be rounded to the
        nearest fixed interval of the DURATION parameter as defined in the ICD
        ranging between 16 fixed values from 0.5 sec to 5 min.
        """
        self.test_action(PFxAction().set_motor_speed(ch, speed, duration=duration))

    def stop_motor(self, ch):
        """
        A convenience wrapper for `PFxAction().stop_motor`

        :param ch: [:obj:`int`] a list of motor channels (1-4)
        """
        self.test_action(PFxAction().stop_motor(ch))

    def light_on(self, ch):
        """
        A convenience wrapper for `PFxAction().light_on`

        :param ch: [:obj:`int`] a list of light channels (1-8)
        """
        self.test_action(PFxAction().light_on(ch))

    def light_off(self, ch):
        """
        A convenience wrapper for `PFxAction().light_off`

        :param ch: [:obj:`int`] a list of light channels (1-8)
        """
        self.test_action(PFxAction().light_off(ch))

    def light_toggle(self, ch):
        """
        A convenience wrapper for `PFxAction().light_toggle`

        :param ch: [:obj:`int`] a list of light channels (1-8)
        """
        self.test_action(PFxAction().light_toggle(ch))

    def set_brightness(self, ch, brightness):
        """
        A convenience wrapper for `PFxAction().set_brightness`

        :param ch: [:obj:`int`] a list of light channels (1-8)
        :param brightness: :obj:`int` brightness (0 - 255 max)
        """
        self.test_action(PFxAction().set_brightness(ch, brightness))

    def combo_light_fx(self, fx, param=[0, 0, 0, 0, 0]):
        """
        A convenience wrapper for `PFxAction().combo_light_fx`

        :param fx: :obj:`int` desired light effect
        :param param: [:obj:`int`] a list of up to 5 light parameters
        """
        self.test_action(PFxAction().combo_light_fx(fx, param=param))

    def light_fx(self, ch, fx, param=[0, 0, 0, 0, 0]):
        """
        A convenience wrapper for `PFxAction().light_fx`

        :param ch: [:obj:`int`] a list of light channels (1-8)
        :param fx: :obj:`int` desired light effect
        :param param: [:obj:`int`] a list of up to 5 light parameters
        """
        self.test_action(PFxAction().light_fx(ch, fx, param=param))

    def sound_fx(self, fx, param=[0, 0], fileID=None):
        """
        A convenience wrapper for `PFxAction().sound_fx`

        :param fx: :obj:`int` desired sound action
        :param param: [:obj:`int`] a list of up to 2 sound parameters
        :param fileID: :obj:`int` file ID of an audio file in the file system
        """
        if fileID is not None:
            fileID = self.file_id_from_str_or_int(fileID)
        self.test_action(PFxAction().sound_fx(fx, param=param, fileID=fileID))

    def play_audio_file(self, fileID):
        """
        A convenience wrapper for `PFxAction().sound_fx`

        :param fileID: :obj:`int` or :obj:`str` file ID or filename of an audio file in the file system
        """
        fileID = self.file_id_from_str_or_int(fileID)
        self.test_action(PFxAction().play_audio_file(fileID=fileID))

    def stop_audio_file(self, fileID):
        """
        A convenience wrapper for `PFxAction().stop_audio_file`

        :param fileID: :obj:`int` or :obj:`str` file ID or filename of an audio file in the file system
        """
        fileID = self.file_id_from_str_or_int(fileID)
        self.test_action(PFxAction().stop_audio_file(fileID=fileID))

    def repeat_audio_file(self, fileID):
        """
        A convenience wrapper for `PFxAction().repeat_audio_file`

        :param fileID: :obj:`int` or :obj:`str` file ID or filename of an audio file in the file system
        """
        fileID = self.file_id_from_str_or_int(fileID)
        self.test_action(PFxAction().repeat_audio_file(fileID=fileID))

    def set_volume(self, volume):
        """
        A convenience wrapper for `PFxAction().set_volume`

        :param volume: :obj:`int` desired audio volume (0 - 100%)
        """
        self.test_action(PFxAction().set_volume(volume))

    def refresh_file_dir(self):
        """
        Reads the PFx Brick file system directory. This includes
        the total storage used as well as the remaining capacity.
        Individual file directory entries are stored in the
        :obj:`PFxBrick.filedir.files` class variable.
        """
        res = cmd_get_free_space(self.dev)
        if res:
            self.filedir.bytesLeft = uint32_toint(res[3:7])
            capacity = uint32_toint(res[7:11])
            self.filedir.bytesUsed = capacity - self.filedir.bytesLeft
        res = cmd_get_num_files(self.dev)
        if res:
            self.filedir.files = []
            self.filedir.numFiles = uint16_toint(res[3:5])
            file_count = 0
            for i in range(PFX_AUDIO_FILES_MAX):
                res = cmd_get_dir_entry(self.dev, i + 1)
                d = PFxFile()
                d.from_bytes(res)
                if d.id < 0xFF:
                    self.filedir.files.append(d)
                    file_count += 1
                if file_count >= self.filedir.numFiles:
                    break

    def put_file(self, fn, fileID=None, show_progress=True):
        """
        Copies a file from the host to the PFx Brick.

        :param fn: :obj:`str` the filename (optionally including the path) of the file to copy
        :param fileID: :obj:`int` the unique file ID to assign the copied file in the file system, can be None for automatic assignment
        :param show_progress: :obj:`boolean` a flag to show the progress bar indicator during transfer.
        """
        if fileID is None:
            self.refresh_file_dir()
            fileID = self.filedir.find_available_file_id()
        if fileID is not None:
            fs_copy_file_to(self, fileID, fn, show_progress)

    def get_file(self, fileID, fn=None, show_progress=True):
        """
        Copies a file from the PFx Brick to the host.

        :param fileID: :obj:`int` or :obj:`str` the file ID or filename of the file to copy
        :param fn: :obj:`str` optional override for the filename when copied into the host
        :param show_progress: :obj:`boolean` a flag to show the progress bar indicator during transfer.
        """
        self.refresh_file_dir()
        fileID = self.file_id_from_str_or_int(fileID)
        f = self.filedir.get_file_dir_entry(fileID)
        fs_copy_file_from(self, f, fn, show_progress)

    def remove_file(self, fileID, silent=False):
        """
        Removes a file from the PFx Brick file system.

        :param fileID: :obj:`int` or :obj:`str` the file ID or filename of the file to remove
        :param silent: :obj:`boolean` if True, warnings about missing files are suppressed
        """
        fileID = self.file_id_from_str_or_int(fileID)
        fs_remove_file(self.dev, fileID, silent=silent)

    def format_fs(self, quick=False):
        """
        Formats the PFx Brick file system, erasing all files.

        :param quick: :obj:`boolean` If True, only occupied sectors are erased. If False, every sector is erased, i.e. a complete format.
        """
        fs_format(self.dev, quick)

    def set_file_attributes(self, fileID, attr, mask=0x7C):
        """
        Sets the upper 8-bit attribute field of a file's directory entry.

        Although the file attribute field is actually 16 bits, this function sets the upper
        8 bits.  The lower 8-bits are reserved for identifying the file type, e.g. text, WAV,
        etc.

        :param fileID: :obj:`int` or :obj:`str` the file ID or filename of the file to remove
        :param attr: :obj:`int` the attribute value to set (only the lower 16)
        :param mask: :obj:`int` an optional bit mask applied to the attributes
        """
        fileID = self.file_id_from_str_or_int(fileID)
        attr = attr & 0xFF
        res = self.send_raw_icd_command(
            [
                PFX_CMD_FILE_DIR,
                PFX_DIR_REQ_SET_ATTR_MASKED_ID,
                fileID & 0xFF,
                0x00,
                attr,
                0x00,
                mask,
            ]
        )

    def rename_file(self, fileID, new_name):
        """
        Renames a file on the file system.

        :param fileID: :obj:`int` or :obj:`str` the file ID or filename of the file to rename
        :param new_name: :obj:`str` new file name to apply (up to 32 characters UTF-8 encoded)
        """
        fileID = self.file_id_from_str_or_int(fileID)
        mb = bytes(new_name, "utf-8")
        name_len = min(len(mb), 32)
        mb = mb[:name_len]
        res = self.send_raw_icd_command(
            [
                PFX_CMD_FILE_DIR,
                PFX_DIR_REQ_RENAME_FILE_ID,
                fileID & 0xFF,
                *mb,
            ]
        )

    def stop_script(self):
        """
        Stops all script execution.
        """
        res = cmd_run_script(self.dev, 0xFF)

    def run_script(self, scriptfile):
        """
        Runs a specified script file on the PFx Brick filesystem.

        A file is identified either by its numeric file ID (0 - 254) or alphanumeric
        filename as a string.

        :param scriptfile: :obj:`int` or :obj:`str` file ID or file name string
        """
        fileID = self.file_id_from_str_or_int(scriptfile)
        res = cmd_run_script(self.dev, fileID)

    def file_id_from_str_or_int(self, filespec):
        """
        Returns a numeric file ID from either a string filename or integer file ID.

        A file is identified on the PFx Brick filesystem primarily by its numeric
        file ID (0 - 254).  A file ID of 255 is an invalid or non-existant file.
        A numeric file ID can be queried using its alphanumeric filename in the
        filesystem directory.  This function performs this lookup if necessary,
        i.e. if a string filename is provided.

        :param filespec: :obj:`int` or :obj:`str` file ID or file name string
        :returns: :obj:`int` numeric file ID from PFx Brick filesystem, or 0xFF if not found
        """
        if isinstance(filespec, int):
            return filespec
        elif isinstance(filespec, str):
            fileid = fs_get_fileid_from_name(self.dev, filespec)
            return fileid
        return 0xFF

    def get_current_state(self):
        """
        Returns the current state of the PFx Brick operating parameters.

        :returns: :obj:`PFxState` a dataclass container with state information
        """
        res = cmd_get_current_state(self.dev)
        self.state.from_bytes(res)
        return self.state

    def get_fs_state(self):
        """
        Returns the current state of the PFx Brick file system.

        :returns: :obj:`PFxFSState` a dataclass container of file system information
        """
        res = self.send_raw_icd_command([PFX_CMD_FILE_GET_FS_STATE])
        self.state.filesys.from_bytes(res)
        return self.state.filesys

    def get_bt_state(self):
        """
        Returns the current state of the PFx Brick Bluetooth radio module.

        :returns: :obj:`PFxBTState` a dataclass container of Bluetooth module information
        """
        res = self.send_raw_icd_command([PFX_CMD_GET_BT_STATUS])
        self.state.bt.from_bytes(res)
        return self.state.bt

    def send_raw_icd_command(self, msg):
        """
        Sends a raw ICD command message represented as a list of bytes.

        :returns: :obj:`bytes` response from the PFx Brick
        """
        res = cmd_raw(self.dev, msg)
        return res



#! /usr/bin/env python3
import argparse

from pfxbrick import *


def full_path(file):
    """Returns the fully expanded path of a file"""
    if "~" in str(file):
        return os.path.expanduser(file)
    return os.path.expanduser(os.path.abspath(file))


def split_path(file):
    """Returns a tuple containing a file's (directory, name.ext)"""
    if os.path.isdir(file):
        return full_path(file), None
    return os.path.split(full_path(file))


def main():
    parser = argparse.ArgumentParser(
        description="copy a file to the PFx Brick from host computer",
        prefix_chars="-+",
    )
    parser.add_argument(
        "file", metavar="file", type=str, help="is the local filename to copy"
    )
    parser.add_argument(
        "dest",
        type=str,
        default=None,
        nargs="?",
        help="is optional file ID instead of next available ID",
    )
    parser.add_argument(
        "-s",
        "--serialno",
        default=None,
        help="Specify PFx Brick with serial number (if more than one connected)",
    )
    args = parser.parse_args()
    argsd = vars(args)

    b = get_one_pfxbrick(argsd["serialno"])
    r = b.open()
    if not r:
        exit()
    b.refresh_file_dir()

    ff = full_path(argsd["file"])
    _, fn = split_path(ff)
    if argsd["dest"] is not None:
        fid = int(argsd["dest"])
        if b.filedir.has_file(fid):
            print("Replacing file %d on PFx Brick..." % (fid))
        else:
            print("Copying file %s as %d to PFx Brick..." % (fn, fid))

    else:
        if b.filedir.has_file(fn):
            fid = b.file_id_from_str_or_int(fn)
            print("Replacing file %s on PFx Brick..." % (fn))
        else:
            fid = b.filedir.find_available_file_id()
            print("Copying file %s to PFx Brick..." % (fn))

    b.put_file(ff, fid)
    if fn.lower().endswith(".pfx"):
        res = b.send_raw_icd_command(
            [PFX_CMD_FILE_DIR, PFX_DIR_REQ_SET_ATTR_ID, fid, 0x30, 0x80]
        )
    b.close()


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
"""
pfxscan - scan for advertising Bluetooth PFx Bricks
"""
import argparse
import asyncio
from datetime import datetime

from rich import print
from rich.console import Console
from rich.table import Table

from pfxbrick import *

console = Console()


async def brick_session(uuid, timeout=15):
    b = PFxBrickBLE(uuid=uuid)
    await b.open(timeout=timeout)
    icd = await b.get_icd_rev()
    await b.get_status()

    if b.status == 0x00:
        name = await b.get_name()
    else:
        name = "Service Mode"
    table = Table(show_header=True, header_style="bold blue")
    bid = "[light_slate_blue]%s [bold cyan]%s" % (b.product_id, b.product_desc)
    table.add_column(bid, width=72)
    table.add_row("Bluetooth UUID        : [bold orange3]%s" % (uuid))
    table.add_row("Serial Number         : [bold cyan]%s" % (b.serial_no))
    table.add_row("ICD Version           : [bold green]%s" % (icd))
    table.add_row(
        "Firmware Version      : [bold green]%s [reset]build [bold green]%s"
        % (b.firmware_ver, b.firmware_build)
    )
    table.add_row(
        "Status                : 0x%02X %s" % (b.status, get_status_str(b.status))
    )
    if b.error:
        table.add_row(
            "Errors                : [red]0x%02X %s" % (b.error, get_error_str(b.error))
        )
    else:
        table.add_row(
            "Errors                : [green]0x%02X %s"
            % (b.error, get_error_str(b.error))
        )
    table.add_row("Name                  : [bold yellow]%s" % (name))
    r = await b.get_rssi()
    table.add_row("RSSI                  : %s dBm" % (r))
    console.print(table)
    await b.close()


def main():
    parser = argparse.ArgumentParser(
        description="Scan for PFx Bricks advertising on Bluetooth"
    )
    parser.add_argument(
        "-s",
        "--scantime",
        default=10,
        help="Time interval (seconds) to scan for advertising PFx Bricks, default=10",
    )
    parser.add_argument(
        "-t",
        "--timeout",
        default=15,
        help="Timeout interval (seconds) to wait while connecting to a PFx Brick, default=15",
    )
    args = parser.parse_args()
    argsd = vars(args)

    loop = asyncio.get_event_loop()
    pfxdevs = []
    with console.status("Scanning...") as status:
        pfxdevs = loop.run_until_complete(
            ble_device_scanner(
                scan_timeout=float(argsd["scantime"]), silent=True, filters=["16 MB"]
            )
        )
    print("Found %d PFx Bricks" % (len(pfxdevs)))
    if len(pfxdevs) > 0:
        for d in pfxdevs:
            with console.status(
                "Connecting to [bold orange3]%s[/]..." % (d.address)
            ) as status:
                loop.run_until_complete(
                    brick_session(d.address, timeout=float(argsd["timeout"]))
                )


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
#
# Copyright (C) 2021  Fx Bricks Inc.
# This file is part of the pfxbrick python module.
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# PFx Brick BLE session helpers


import asyncio
import logging
import sys

from bleak import BleakClient, BleakScanner

from pfxbrick import *
from pfxbrick.pfxexceptions import *
from pfxbrick.pfxhelpers import *
from pfxbrick.pfxmsg import *

DEV_INFO_UUID = "%x" % (PFX_BLE_GATT_DEV_INFO_UUID)
DEV_SN_UUID = "%x" % (PFX_BLE_GATT_DEV_SN_UUID)
MAX_RETRIES = 200
RX_RETRY_TIME = 0.05


async def ble_device_scanner(
    scan_time=3.0,
    min_devices=1,
    scan_timeout=30.0,
    filters=None,
    silent=False,
    verbose=False,
):
    """
    Performs a Bluetooth scan for available peripheral devices which advertise
    themselves as PFx Bricks.

    This coroutine will search for the required number of devices in scan_time
    chunks up until the scan_timeout interval has elapsed.

    :param scan_time: :obj:`float` scan time interval to look for devices
    :param min_devices: :obj:`int` minimum number of devices to look for before returning
    :param scan_timeout: :obj:`float` timeout interval for finding the required min_devices
    :param filters: :obj:`list` or `str` optional device name filters, e.g. "16 MB"
    :param silent: :obj:`boolean` a flag to disable printing of status
    :param verbose: :obj:`boolean` a flag to print a verbose list of advertising devices

    :returns: [:obj:`BLEDevice`] a list of PFx Brick device objects described in a Bleak BLEDevice class.
    """

    def _filter_device(dc):
        if filters is not None:
            for f in filters:
                if f in dc.name:
                    return True
            return False
        return True

    pfxdevs = []
    total_scan_time = 0
    if filters is not None:
        if isinstance(filters, str):
            filters = [filters]
    while len(pfxdevs) < min_devices:
        if not silent:
            print("Scanning...")
        async with BleakScanner() as scanner:
            await asyncio.sleep(scan_time)
            total_scan_time += scan_time
            devices = scanner.discovered_devices
        if not silent and len(devices) > 0:
            print("Found %d advertising devices" % (len(devices)))
            if verbose:
                for i, d in enumerate(devices):
                    s = '%2d. UUID=%-36s "%s"' % (i + 1, d.address, d.name)
                    print(s)
        for d in devices:
            if "PFx Brick" in d.name:
                if _filter_device(d):
                    if not silent:
                        print("Found %s UUID=%-36s" % (d.name, d.address))
                    pfxdevs.append(d)
        if total_scan_time > scan_timeout:
            break
    return pfxdevs


async def find_ble_pfxbricks(devices, connect_interval=5.0, timeout=30.0, silent=False):
    """
    Resolves a list of scanned candidate Bluetooth devices into valid PFx Brick devices.

    :param devices: [:obj:`BLEDevice`] a list of candidate PFx Brick device objects
    :param connect_interval: :obj:`float` time interval to wait for connection
    :param timeout: :obj:`float` timeout interval for attempting a device connection
    :param silent: :obj:`boolean` a flag to disable printing of status

    :returns: [:obj:`dict`] a list of dictionary objects for each PFx Brick verified by connection. The dictionary contains keys for "address", "serial_no" and "name".
    """

    async def connect_device(device, connect_timeout):
        async with BleakClient(device.address, timeout=connect_timeout) as client:
            sn = None
            # only continue if we have a connection to the device
            if client.is_connected:
                rssi = device.rssi
                # get all the available services from the device
                services = await client.get_services()
                for service in services:
                    if DEV_INFO_UUID.lower() in service.uuid:
                        for char in service.characteristics:
                            if DEV_SN_UUID in char.uuid:
                                sn = bytes(await client.read_gatt_char(char.uuid))
                                return sn, rssi

    pfxbricks = []
    for d in devices:
        sn = None
        connected = False
        total_connect_time = 0
        while not connected and total_connect_time < timeout:
            try:
                if not silent:
                    print('Discovering device %s "%s"...' % (d.address, d.name))
                sn, rssi = await connect_device(d, connect_timeout=connect_interval)
                connected = True
            except:
                total_connect_time += connect_interval
        if sn is not None:
            if not silent:
                fmt = 'Found "%s" S/N=%s UUID=%s'
                if rssi is not None:
                    fmt = fmt + " RSSI=%d dBm"
                    print(fmt % (d.name, str(sn, encoding="utf8"), d.address, rssi))
                else:
                    print(fmt % (d.name, str(sn, encoding="utf8"), d.address))
            pfxbricks.append({"address": d.address, "serial_no": sn, "name": d.name})

    return pfxbricks


class PFxBrickBLE(PFxBrick):
    """
    Inherited sub-class of a :obj:`PFxBrick` object class.

    This sub-class of :obj:`PFxBrick` is necessary for communicating with a
    PFx Brick via Bluetooth LE.  Since the Bluetooth communication stack is
    based on the Bleak python module, it runs in an asynchronous context using
    python's async/await mechanisms.  Since this requires co-routines instead
    of conventional function methods, this sub-class reimplements many of
    the parent :obj:`PFxBrick` class methods as asyncronous co-routines.

    Almost all class functionality is more or less the same as the USB based
    PFxBrick class and fortunately some of the utility methods can be
    reused.

    You can initialize a `PFxBrickBLE` object instance with either a dictionary
    or a UUID string that specifies which PFx Brick to connect with.

    The dictionary describing the desired PFx Brick peripheral device
    is as follows:

    "address":  hardware address of the PFx Brick obtained by a Bluetooth device scan
    "serial_no":  optional serial number of the PFx Brick obtained by a device scan
    "name": optional name of the PFx Brick device obtained by a device scan

    Only the "address" key is mandatory, the other keys are provided if desired.

    Alternatively, you can simply initialize the PFxBrickBLE class using a
    UUID string like the following example:

    brick = PFxBrickBLE(uuid="059930E2-BE75-48A4-B193-3AD3F67246E4")

    Unless the Bluetooth hardware address of the PFx Brick is known in advance,
    then it must be obtained by performing a Bluetooth peripheral device scan to
    see which Bluetooth devices are currently advertising availability.  The
    Bluetooth hardware address is operating system dependent and must be provided
    in a UUID form that is compatible with your OS.

    for Windows and Linux this is typically in the form of "24:71:89:cc:09:05"
    and on macOS it is in the form of "B9EA5233-37EF-4DD6-87A8-2A875E821C46"

    Attributes:
        dev (:obj:`device`): a device handle which is reference to self

        is_open (:obj:`boolean`): a flag indicating connected session status

        client (:obj:`BleakClient`): Bleak BLE client object reference

        callback_audio_done (:obj:`func`): a function callback reference in response to a `PFX_NOTIFICATION_AUDIO_PLAY_DONE` notification. Must have the call signature `func(fileid, filename)`

        callback_audio_play (:obj:`func`): a function callback reference in response to a `PFX_NOTIFICATION_AUDIO_PLAY` notification. Must have the call signature `func(fileid, filename)`

        callback_motora_stop (:obj:`func`): a function callback reference in response to a `PFX_NOTIFICATION_MOTORA_STOP` notification. Must have the call signature `func()`

        callback_motora_speed (:obj:`func`): a function callback reference in response to a `PFX_NOTIFICATION_MOTORA_CURR_SPD` notification. Must have the call signature `func(speed)`

        callback_motorb_stop (:obj:`func`): a function callback reference in response to a `PFX_NOTIFICATION_MOTORB_STOP` notification. Must have the call signature `func()`

        callback_motorb_speed (:obj:`func`): a function callback reference in response to a `PFX_NOTIFICATION_MOTORB_CURR_SPD` notification. Must have the call signature `func(speed)`

    :param dev_dict: :obj:`dict` a dictionary describing the PFx Brick device to connect. Must have the key "address" with the Bluetooth MAC address of the PFx Brick. Optional keys "name" and "serial_no" can be provided.
    :param uuid: :obj:`str` a string representing the UUID/address of the PFx Brick device. This is an alternative to using the `dev_dict` argument to specify the PFx Brick.
    :param debug: :obj:`boolean` a flag to enable low level debug logging of Bluetooth session activity
    """

    def __init__(self, dev_dict=None, uuid=None, debug=False):
        super().__init__()
        if dev_dict is not None:
            if "address" not in dev_dict:
                raise BLEDeviceMissingAddressException()
            else:
                self.ble_address = dev_dict["address"]
            if "serial_no" in dev_dict:
                self.serial_no = dev_dict["serial_no"]
            if "name" in dev_dict:
                self.usb_prod_str = dev_dict["name"]
        elif uuid is not None:
            self.ble_address = uuid
        else:
            raise BLEDeviceMissingAddressException()
        self.client = None
        self.is_open = False
        self.dev = None
        self.callback_audio_done = None
        self.callback_audio_play = None
        self.callback_motora_stop = None
        self.callback_motora_speed = None
        self.callback_motorb_stop = None
        self.callback_motorb_speed = None
        self._disconnect_flag = False
        self._rxbuff = None
        self._log = logging.getLogger(str(self.__class__))
        if debug:
            self._log.setLevel(logging.DEBUG)
            h = logging.StreamHandler(sys.stdout)
            h.setLevel(logging.DEBUG)
            self._log.addHandler(h)

    async def open(self, timeout=10):
        """
        Opens a BLE communication session with a PFx Brick.

        This method is called after this instance has been initialized with a valid
        Bluetooth address.

        :param timeout: :obj:`float` timeout interval (seconds) to wait for an open connection
        """
        self.client = BleakClient(
            self.ble_address, disconnected_callback=self._disconnected_callback
        )
        self.is_open = await self.client.connect(timeout=timeout)
        if self.is_open:
            self._log.info("Connected to PFx Brick %s" % (self.ble_address))
            await self.client.start_notify(PFX_BLE_GATT_UART_RX_UUID, self._rx_callback)
            self.dev = self
            self.usb_manu_str = "Fx Bricks"
        else:
            self._log.error("Timeout connecting to %s" % (self.ble_address))
            raise BLEConnectTimeoutException()

    async def close(self):
        """
        Closes a BLE communication session with a PFx Brick.
        """
        if self.is_open and self.client is not None:
            self._disconnect_flag = True
            await self.client.disconnect()
            self._log.info("Connection closed with PFx Brick %s" % (self.ble_address))

    async def get_rssi(self):
        """
        Get a recent measurement of RSSI (received signal strength)

        :returns: :obj:`int` or None of signal strength in dBm
        """
        rssi = None
        if self.is_open:
            rssi = await self.client.get_rssi()
        return rssi

    def _disconnected_callback(self, client):
        """
        BLE disconnection event handler.
        This is always called by the Bleak BLE API since this callback is registered
        in the :obj:`open` method.
        """
        if not self._disconnect_flag:
            self._log.warning(
                "Unexpected disconnection from PFx Brick %s" % (self.ble_address)
            )
            raise BLEDeviceDisconnectedException()

    def _process_notification(self, msg):
        if msg[0] == PFX_MSG_NOTIFICATION:
            if msg[1] == PFX_NOTIFICATION_AUDIO_PLAY_DONE:
                if self.callback_audio_done is not None:
                    fn = self.filedir.get_file_dir_entry(msg[2])
                    self.callback_audio_done(msg[2], fn.name)
            if msg[1] == PFX_NOTIFICATION_AUDIO_PLAY:
                if self.callback_audio_play is not None:
                    fn = self.filedir.get_file_dir_entry(msg[2])
                    self.callback_audio_play(msg[2], fn.name)
            if msg[1] == PFX_NOTIFICATION_MOTORA_CURR_SPD:
                if self.callback_motora_speed is not None:
                    self.callback_motora_speed(int8_toint(msg[2]))
            if msg[1] == PFX_NOTIFICATION_MOTORA_STOP:
                if self.callback_motora_stop is not None:
                    self.callback_motora_stop()
            if msg[1] == PFX_NOTIFICATION_MOTORB_CURR_SPD:
                if self.callback_motorb_speed is not None:
                    self.callback_motorb_speed(int8_toint(msg[2]))
            if msg[1] == PFX_NOTIFICATION_MOTORB_STOP:
                if self.callback_motorb_stop is not None:
                    self.callback_motorb_stop()

    def _rx_callback(self, sender, data):
        self._log.info("Rx Data: %s" % (data))
        self._rxbuff = data
        # look for notificaitons in the received buffer
        # and activate callbacks if required
        last_idx = -3
        for i, b in enumerate(self._rxbuff):
            if b == PFX_MSG_NOTIFICATION and i > last_idx + 2:
                if (i + 2) < len(self._rxbuff):
                    self._process_notification(self._rxbuff[i : i + 3])
                    last_idx = i

    async def _tx_msg(self, msg):
        self._rxbuff = []
        msg_type = int(0x80 | msg[3])
        chunks = [msg[i : i + 20] for i in range(0, len(msg), 20)]
        for chunk in chunks:
            await self.client.write_gatt_char(PFX_BLE_GATT_UART_TX_UUID, chunk)
            self._log.info("Tx Data: %s" % (chunk))
        retries = 0
        while len(self._rxbuff) == 0 and retries < MAX_RETRIES:
            await asyncio.sleep(RX_RETRY_TIME)
            retries += 1
        if retries >= MAX_RETRIES:
            self._log.error(
                "Timeout waiting for response from PFx Brick %s" % (self.ble_address)
            )
            await self.close()
            raise ResponseTimeoutException()
        if len(self._rxbuff) > 0:
            if self._rxbuff[0] != msg_type and self._rxbuff[0] != PFX_MSG_NOTIFICATION:
                self._log.error(
                    "Invalid response from PFx Brick %s" % (self.ble_address)
                )
                await self.close()
                raise InvalidResponseException()

    async def ble_transaction(self, msg):
        """
        Wraps and sends an ICD message via Bluetooth and waits for and returns a
        corresponding response from the PFx Brick.

        :param msg: [:obj:`int`] ICD message to send as an integeter list of bytes
        :returns: [:obj:`int`] returned message in a byte array list
        """
        tx = bytearray()
        # wrap the message with the required prefix/suffix delimiters [[[ ]]]
        tx.extend([0x5B, 0x5B, 0x5B])
        tx.extend(msg)
        tx.extend([0x5D, 0x5D, 0x5D])
        await self._tx_msg(tx)
        return self._rxbuff

    async def get_icd_rev(self, silent=False):
        """
        Requests the version of Interface Control Document (ICD)
        the connected PFx Brick supports using the `PFX_CMD_GET_ICD_REV`
        ICD message.  The resulting version number is stored in
        this class and also returned.

        :param boolean silent: flag to optionally silence the status LED blink

        :returns: :obj:`str` ICD version, e.g. "3.38"
        """
        res = await cmd_get_icd_rev(self.dev, silent)
        self.icd_rev = uint16_tover(res[1], res[2])
        self.config.icd_rev = self.icd_rev
        return self.icd_rev

    async def get_status(self):
        """
        Requests the top level operational status of the PFx Brick
        using the `PFX_CMD_GET_STATUS ICD` message.  The resulting
        status data is stored in this class and can be queried
        with typical class member access methods or the print_status method.
        """
        res = await cmd_get_status(self.dev)
        if res:
            self.status = res[1]
            self.error = res[2]
            self.product_id = uint16_tostr(res[7], res[8])
            self.serial_no = uint32_tostr(res[9], res[10], res[11], res[12])
            self.product_desc = bytes(res[13:37]).decode("utf-8")
            self.firmware_ver = uint16_tover(res[37], res[38])
            self.firmware_build = uint16_tostr(res[39], res[40])

    async def get_config(self):
        """
        Retrieves configuration settings from the PFx Brick using
        the `PFX_CMD_GET_CONFIG ICD` message. The configuration data
        is stored in the :obj:`PFxBrick.config` class member variable.
        """
        res = await cmd_get_config(self.dev)
        if res:
            self.config.from_bytes(res)

    async def set_config(self):
        """
        Writes the contents of the PFxConfig data structure class to
        the PFx Brick using the `PFX_CMD_SET_CONFIG` ICD message.

        It is recommended that the configuration be read from the
        PFx Brick (using get_config) before any changes are made to
        the configuration and written back. This ensures that any
        configuration settings which are not desired to be changed
        are left in the same state.
        """
        res = await cmd_set_config(self.dev, self.config.to_bytes())

    async def reset_factory_config(self):
        """
        Resets the PFx Brick configuration settings to factory defaults.
        """
        res = await cmd_set_factory_defaults(self.dev)

    async def get_name(self):
        """
        Retrieves the user defined name of the PFx Brick using
        the `PFX_CMD_GET_NAME ICD` message. The name is stored in
        the name class variable as a UTF-8 string.

        :returns: :obj:`str` user defined name
        """
        res = await cmd_get_name(self.dev)
        if res:
            self.name = safe_unicode_str(res[1:25])
        return self.name

    async def set_name(self, name):
        """
        Sets the user defined name of the PFx Brick using the
        `PFX_CMD_SET_NAME` ICD message.

        :param name: :obj:`str` new name to set (up to 24 character bytes, UTF-8)
        """
        res = await cmd_set_name(self.dev, name)

    async def get_action_by_address(self, address):
        """
        Retrieves a stored action indexed by address rather than a
        combination of eventID and IR channel.  The address is converted into a
        [eventID, IR channel] pair and the get_action method is
        called with this function as a convenient wrapper.

        :param address: :obj:`int` event/action LUT address (0 - 0x7F)
        :returns: :obj:`PFxAction` class filled with retrieved LUT data
        """
        if address > EVT_LUT_MAX:
            self._log.warning(
                "Requested action at address %02X is out of range" % (address)
            )
            return None
        else:
            evt, ch = address_to_evtch(address)
            a = await self.get_action(evt, ch)
            return a

    async def get_action(self, evtID, ch):
        """
        Retrieves the stored action associated with a particular
        [eventID / IR channel] event. The eventID and channel value
        form a composite address pointer into the event/action LUT
        in the PFx Brick. The address to the LUT is formed as:

        Address[5:2] = event ID
        Address[1:0] = channel

        :param evtID: :obj:`int` event ID LUT address component (0 - 0x20)
        :param channel: :obj:`int` channel index LUT address component (0 - 3)
        :returns: :obj:`PFxAction` class filled with retrieved LUT data
        """
        if ch > 3 or evtID > EVT_ID_MAX:
            self._log.warning(
                "Requested action (id=%02X, ch=%02X) is out of range" % (evtID, ch)
            )
            return None
        else:
            res = await cmd_get_event_action(self.dev, evtID, ch)
            action = PFxAction()
            if res:
                action.from_bytes(res)
            return action

    async def set_action_by_address(self, address, action):
        """
        Sets a new stored action in the event/action LUT at the
        address specified. The address is converted into a
        [eventID, IR channel] pair and the set_action method is
        called with this function as a convenient wrapper.

        :param address: :obj:`int` event/action LUT address (0 - 0x7F)
        :param action: :obj:`PFxAction` action data structure class
        """
        if address > EVT_LUT_MAX:
            self._log.warning(
                "Requested action at address %02X is out of range" % (address)
            )
            return None
        else:
            evt, ch = address_to_evtch(address)
            await self.set_action(evt, ch, action)

    async def set_action(self, evtID, ch, action):
        """
        Sets a new stored action associated with a particular
        [eventID / IR channel] event. The eventID and channel value
        form a composite address pointer into the event/action LUT
        in the PFx Brick. The address to the LUT is formed as:

        Address[5:2] = event ID
        Address[1:0] = channel

        :param evtID: :obj:`int` event ID LUT address component (0 - 0x20)
        :param ch: :obj:`int` channel index LUT address component (0 - 3)
        :param action: :obj:`PFxAction` action data structure class
        """
        if ch > 3 or evtID > EVT_ID_MAX:
            self._log.warning(
                "Requested action (id=%02X, ch=%02X) is out of range" % (evtID, ch)
            )
            return None
        else:
            res = await cmd_set_event_action(self.dev, evtID, ch, action.to_bytes())

    async def test_action(self, action):
        """
        Executes a passed action data structure. This function is
        used to "test" actions to see how they behave. The passed
        action is not stored in the event/action LUT.

        :param action: :obj:`PFxAction` action data structure class
        """
        res = await cmd_test_action(self.dev, action.to_bytes())

    async def clear_action_by_address(self, address):
        """
        Clears a stored action in the event/action LUT at the
        address specified. The address is converted into a
        [eventID, IR channel] pair and the set_action method is
        called with this function as a convenient wrapper.

        :param address: :obj:`int` event/action LUT address (0 - 0x7F)
                        :obj:`list,tuple,range` specify a list or range of addresses
        """
        if isinstance(address, (list, tuple)):
            addresses = address
        elif isinstance(address, range):
            addresses = [x for x in range]
        else:
            addresses = [address]
        for a in addresses:
            if a > EVT_LUT_MAX:
                print("Requested action at address %02X is out of range" % (a))
                return None
            else:
                evt, ch = address_to_evtch(a)
                await self.clear_action(evt, ch)

    async def clear_action(self, evtID, ch):
        """
        Clears a stored action associated with a particular
        [eventID / IR channel] event. The eventID and channel value
        form a composite address pointer into the event/action LUT
        in the PFx Brick. The address to the LUT is formed as:

        Address[5:2] = event ID
        Address[1:0] = channel

        :param evtID: :obj:`int` event ID LUT address component (0 - 0x20)
        :param ch: :obj:`int` channel index LUT address component (0 - 3)
        """
        if ch > 3 or evtID > EVT_ID_MAX:
            print("Requested action (id=%02X, ch=%02X) is out of range" % (evtID, ch))
            return None
        else:
            # set to an empty PFxAction to clear
            res = await cmd_set_event_action(
                self.dev, evtID, ch, PFxAction().to_bytes()
            )

    async def find_startup_action(self, lightfx=None, soundfx=None, motorfx=None):
        raise NotImplementedError("PFx Brick method not supported over Bluetooth")

    async def set_motor_speed(self, ch, speed, duration=None):
        """
        A convenience wrapper for `PFxAction().set_motor_speed`

        :param ch: [:obj:`int`] a list of motor channels (1-4)
        :param speed: :obj:`int` desired motor speed (-100 to +100)
        :param duration: :obj:`float` optional duration (in seconds) to run motor, runs indefinitely if not specified

        If the duration value is specified, it represents the desired motor
        run time in seconds. Note that this value will be rounded to the
        nearest fixed interval of the DURATION parameter as defined in the ICD
        ranging between 16 fixed values from 0.5 sec to 5 min.
        """
        await self.test_action(
            PFxAction().set_motor_speed(ch, speed, duration=duration)
        )

    async def stop_motor(self, ch):
        """
        A convenience wrapper for `PFxAction().stop_motor`

        :param ch: [:obj:`int`] a list of motor channels (1-4)
        """
        await self.test_action(PFxAction().stop_motor(ch))

    async def light_on(self, ch):
        """
        A convenience wrapper for `PFxAction().light_on`

        :param ch: [:obj:`int`] a list of light channels (1-8)
        """
        await self.test_action(PFxAction().light_on(ch))

    async def light_off(self, ch):
        """
        A convenience wrapper for `PFxAction().light_off`

        :param ch: [:obj:`int`] a list of light channels (1-8)
        """
        await self.test_action(PFxAction().light_off(ch))

    async def light_toggle(self, ch):
        """
        A convenience wrapper `for PFxAction().light_toggle`

        :param ch: [:obj:`int`] a list of light channels (1-8)
        """
        await self.test_action(PFxAction().light_toggle(ch))

    async def set_brightness(self, ch, brightness):
        """
        A convenience wrapper for `PFxAction().set_brightness`

        :param ch: [:obj:`int`] a list of light channels (1-8)
        :param brightness: :obj:`int` brightness (0 - 255 max)
        """
        await self.test_action(PFxAction().set_brightness(ch, brightness))

    async def combo_light_fx(self, fx, param=[0, 0, 0, 0, 0]):
        """
        A convenience wrapper for `PFxAction().combo_light_fx`

        :param fx: :obj:`int` desired light effect
        :param param: [:obj:`int`] a list of up to 5 light parameters
        """
        await self.test_action(PFxAction().combo_light_fx(fx, param=param))

    async def light_fx(self, ch, fx, param=[0, 0, 0, 0, 0]):
        """
        A convenience wrapper for `PFxAction().light_fx`

        :param ch: [:obj:`int`] a list of light channels (1-8)
        :param fx: :obj:`int` desired light effect
        :param param: [:obj:`int`] a list of up to 5 light parameters
        """
        await self.test_action(PFxAction().light_fx(ch, fx, param=param))

    async def sound_fx(self, fx, param=[0, 0], fileID=None):
        """
        A convenience wrapper for `PFxAction().sound_fx`

        :param fx: :obj:`int` desired sound action
        :param param: [:obj:`int`] a list of up to 2 sound parameters
        :param fileID: :obj:`int` file ID of an audio file in the file system
        """
        if fileID is not None:
            fileID = await self.file_id_from_str_or_int(fileID)
        await self.test_action(PFxAction().sound_fx(fx, param=param, fileID=fileID))

    async def play_audio_file(self, fileID):
        """
        A convenience wrapper for `PFxAction().sound_fx`

        :param fileID: :obj:`int` or :obj:`str` file ID or filename of an audio file in the file system
        """
        fileID = await self.file_id_from_str_or_int(fileID)
        await self.test_action(PFxAction().play_audio_file(fileID=fileID))

    async def stop_audio_file(self, fileID):
        """
        A convenience wrapper for `PFxAction().stop_audio_file`

        :param fileID: :obj:`int` or :obj:`str` file ID or filename of an audio file in the file system
        """
        fileID = await self.file_id_from_str_or_int(fileID)
        await self.test_action(PFxAction().stop_audio_file(fileID=fileID))

    async def repeat_audio_file(self, fileID):
        """
        A convenience wrapper for `PFxAction().repeat_audio_file`

        :param fileID: :obj:`int` or :obj:`str` file ID or filename of an audio file in the file system
        """
        fileID = await self.file_id_from_str_or_int(fileID)
        await self.test_action(PFxAction().repeat_audio_file(fileID=fileID))

    async def set_volume(self, volume):
        """
        A convenience wrapper for `PFxAction().set_volume`

        :param volume: :obj:`int` desired audio volume (0 - 100%)
        """
        await self.test_action(PFxAction().set_volume(volume))

    async def refresh_file_dir(self):
        """
        Reads the PFx Brick file system directory. This includes
        the total storage used as well as the remaining capacity.
        Individual file directory entries are stored in the
        :obj:`PFxBrick.filedir.files` class variable.
        """
        res = await cmd_get_free_space(self.dev)
        if res:
            self.filedir.bytesLeft = uint32_toint(res[3:7])
            capacity = uint32_toint(res[7:11])
            self.filedir.bytesUsed = capacity - self.filedir.bytesLeft
        res = await cmd_get_num_files(self.dev)
        if res:
            self.filedir.files = []
            self.filedir.numFiles = uint16_toint(res[3:5])
            file_count = 0
            for i in range(PFX_AUDIO_FILES_MAX):
                res = await cmd_get_dir_entry(self.dev, i + 1)
                d = PFxFile()
                d.from_bytes(res)
                if d.id < 0xFF:
                    self.filedir.files.append(d)
                    file_count += 1
                if file_count >= self.filedir.numFiles:
                    break

    async def put_file(self, fileID, fn, show_progress=True):
        """
        PFx Brick file system operations not supported over Bluetooth
        raises :obj:`NotImplementedError`
        """
        raise NotImplementedError(
            "PFx Brick file system operations not supported over Bluetooth"
        )

    async def get_file(self, fileID, fn=None, show_progress=True):
        """
        PFx Brick file system operations not supported over Bluetooth
        raises :obj:`NotImplementedError`
        """
        raise NotImplementedError(
            "PFx Brick file system operations not supported over Bluetooth"
        )

    async def remove_file(self, fileID):
        """
        PFx Brick file system operations not supported over Bluetooth
        raises :obj:`NotImplementedError`
        """
        raise NotImplementedError(
            "PFx Brick file system operations not supported over Bluetooth"
        )

    async def format_fs(self, quick=False):
        """
        PFx Brick file system operations not supported over Bluetooth
        raises :obj:`NotImplementedError`
        """
        raise NotImplementedError(
            "PFx Brick file system operations not supported over Bluetooth"
        )

    async def set_file_attributes(self, fileID, attr, mask=0x7C):
        """
        PFx Brick file system operations not supported over Bluetooth
        raises :obj:`NotImplementedError`
        """
        raise NotImplementedError(
            "PFx Brick file system operations not supported over Bluetooth"
        )

    async def rename_file(self, fileID, new_name):
        """
        PFx Brick file system operations not supported over Bluetooth
        raises :obj:`NotImplementedError`
        """
        raise NotImplementedError(
            "PFx Brick file system operations not supported over Bluetooth"
        )

    async def stop_script(self):
        """
        Stops all script execution.
        """
        res = await cmd_run_script(self.dev, 0xFF)

    async def run_script(self, scriptfile):
        """
        Runs a specified script file on the PFx Brick filesystem.

        A file is identified either by its numeric file ID (0 - 254) or alphanumeric
        filename as a string.

        :param scriptfile: :obj:`int` or :obj:`str` file ID or file name string
        """
        fileid = await self.file_id_from_str_or_int(scriptfile)
        res = await cmd_run_script(self.dev, fileid)

    async def file_id_from_str_or_int(self, filespec):
        """
        Returns a numeric file ID from either a string filename or integer file ID.

        A file is identified on the PFx Brick filesystem primarily by its numeric
        file ID (0 - 254).  A file ID of 255 is an invalid or non-existant file.
        A numeric file ID can be queried using its alphanumeric filename in the
        filesystem directory.  This function performs this lookup if necessary,
        i.e. if a string filename is provided.

        :param filespec: :obj:`int` or :obj:`str` file ID or file name string
        :returns: :obj:`int` numeric file ID from PFx Brick filesystem, or 0xFF if not found
        """
        if isinstance(filespec, int):
            return filespec
        elif isinstance(filespec, str):
            fileid = 0xFF
            fb = bytes(filespec, "utf-8")
            p = [len(fb)]
            p.extend(fb)
            res = await cmd_file_dir(self.dev, PFX_DIR_REQ_GET_NAMED_FILE_ID, p)
            if len(res) >= 3 and not res[2] == PFX_ERR_FILE_NOT_FOUND:
                fileid = int(res[2])
            return fileid
        return 0xFF

    async def get_current_state(self):
        """
        Returns the current state of the PFx Brick operating parameters.

        :returns: :obj:`PFxState` a dataclass container with state information
        """
        res = await cmd_get_current_state(self.dev)
        self.state.from_bytes(res)
        return self.state

    async def get_fs_state(self):
        """
        PFx Brick operation not supported over Bluetooth
        raises :obj:`NotImplementedError`
        """
        raise NotImplementedError("PFx Brick method not supported over Bluetooth")

    async def get_bt_state(self):
        """
        PFx Brick operation not supported over Bluetooth
        raises :obj:`NotImplementedError`
        """
        raise NotImplementedError("PFx Brick method not supported over Bluetooth")

    async def send_raw_icd_command(self, msg):
        """
        Sends a raw ICD command message represented as a list of bytes.

        :returns: :obj:`bytes` response from the PFx Brick
        """
        res = await cmd_raw(self.dev, msg)
        return res

    async def set_notifications(self, events):
        """
        Enables user selected notifications to be sent asynchronously from the PFx Brick.

        :param events: :obj:`int` a bitwise OR of notification flags:

        - :obj:`PFX_NOTIFICATION_AUDIO_PLAY_DONE = 0x01`
        - :obj:`PFX_NOTIFICATION_AUDIO_PLAY = 0x02`
        - :obj:`PFX_NOTIFICATION_MOTORA_CURR_SPD = 0x04`
        - :obj:`PFX_NOTIFICATION_MOTORA_STOP = 0x08`
        - :obj:`PFX_NOTIFICATION_MOTORB_CURR_SPD = 0x10`
        - :obj:`PFX_NOTIFICATION_MOTORB_STOP = 0x20`
        - :obj:`PFX_NOTIFICATION_TO_USB = 0x80`
        - :obj:`PFX_NOTIFICATION_TO_BLE = 0x40`

        Note that :obj:`PFX_NOTIFICATION_TO_BLE` is automatically set and does not need to be specified.
        """

        # if notifications are configured for audio events, refresh file directory
        # so that we can resolve file ID numbers to filenames
        if (
            events & PFX_NOTIFICATION_AUDIO_PLAY
            or events & PFX_NOTIFICATION_AUDIO_PLAY_DONE
        ):
            await self.refresh_file_dir()
        res = await cmd_set_notifications(self.dev, PFX_NOTIFICATION_TO_BLE | events)

    async def disable_notifications(self):
        """
        Disables asynchronous notifications sent from the PFx Brick.
        """
        res = await cmd_set_notifications(self.dev, 0)



#! /usr/bin/env python3
#
# Copyright (C) 2018  Fx Bricks Inc.
# This file is part of the pfxbrick python module.
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# PFx Brick message helpers


import hid

from .pfx import *
from .pfxexceptions import InvalidResponseException
from .pfxhelpers import uint32_to_bytes


def usb_transaction(hdev, msg):
    # enforce non-numbered report pre-pending and report length
    # This ensures consistent operation on Windows, macOS, etc.
    # since Windows insists on matched report length/buffer size
    # and for all non-numbered reports to start with 0
    msglen = len(msg)
    buf = [0]
    buf.extend(msg)
    buf.extend([0] * (64 - msglen))
    hdev.write(buf)
    if not msg[0] == PFX_CMD_FILE_WRITE_FAST:
        res = hdev.read(64)
    else:
        res = [msg[0] | 0x80]
        res.extend([0] * 63)
    if res:
        if res[0] == msg[0] | 0x80:
            return res
        else:
            raise InvalidResponseException()
    return 0


def msg_transaction(hdev, msg):
    if isinstance(hdev, hid.device):
        return usb_transaction(hdev, msg)
    else:
        return hdev.ble_transaction(msg)


def cmd_get_icd_rev(hdev, silent=False):
    msg = [
        PFX_CMD_GET_ICD_REV,
        PFX_GET_ICD_BYTE0,
        PFX_GET_ICD_BYTE1,
        PFX_GET_ICD_BYTE2,
        int(silent),
    ]
    return msg_transaction(hdev, msg)


def cmd_get_status(hdev):
    msg = [
        PFX_CMD_GET_STATUS,
        PFX_STATUS_BYTE0,
        PFX_STATUS_BYTE1,
        PFX_STATUS_BYTE2,
        PFX_STATUS_BYTE3,
        PFX_STATUS_BYTE4,
        PFX_STATUS_BYTE5,
        PFX_STATUS_BYTE6,
    ]
    return msg_transaction(hdev, msg)


def cmd_get_config(hdev):
    msg = [PFX_CMD_GET_CONFIG]
    return msg_transaction(hdev, msg)


def cmd_set_config(hdev, cfgbytes):
    msg = [PFX_CMD_SET_CONFIG]
    msg.extend(cfgbytes)
    return msg_transaction(hdev, msg)


def cmd_get_name(hdev):
    msg = [PFX_CMD_GET_NAME]
    return msg_transaction(hdev, msg)


def cmd_set_name(hdev, name):
    msg = [PFX_CMD_SET_NAME]
    mb = bytes(name, "utf-8")
    for x in mb:
        msg.append(x)
    for i in range(PFX_NAME_MAX - len(mb)):
        msg.append(0)
    return msg_transaction(hdev, msg)


def cmd_get_event_action(hdev, evtID, ch):
    msg = [PFX_CMD_GET_EVENT_ACTION, evtID, ch]
    return msg_transaction(hdev, msg)


def cmd_set_event_action(hdev, evtID, ch, action):
    msg = [PFX_CMD_SET_EVENT_ACTION, evtID, ch]
    msg.extend(action)
    return msg_transaction(hdev, msg)


def cmd_test_action(hdev, action):
    msg = [PFX_CMD_TEST_ACTION]
    msg.extend(action)
    return msg_transaction(hdev, msg)


def cmd_get_dir_entry(hdev, idx):
    msg = [PFX_CMD_FILE_DIR, PFX_DIR_REQ_GET_DIR_ENTRY_IDX, idx]
    return msg_transaction(hdev, msg)


def cmd_get_num_files(hdev):
    msg = [PFX_CMD_FILE_DIR, PFX_DIR_REQ_GET_FILE_COUNT]
    return msg_transaction(hdev, msg)


def cmd_get_free_space(hdev):
    msg = [PFX_CMD_FILE_DIR, PFX_DIR_REQ_GET_FREE_SPACE]
    return msg_transaction(hdev, msg)


def cmd_set_factory_defaults(hdev):
    msg = [
        PFX_CMD_SET_FACTORY_DEFAULTS,
        PFX_RESET_BYTE0,
        PFX_RESET_BYTE1,
        PFX_RESET_BYTE2,
        PFX_RESET_BYTE3,
        PFX_RESET_BYTE4,
        PFX_RESET_BYTE5,
        PFX_RESET_BYTE6,
    ]
    return msg_transaction(hdev, msg)


def cmd_set_notifications(hdev, flags):
    msg = [PFX_CMD_SET_NOTIFICATIONS, flags]
    return msg_transaction(hdev, msg)


def cmd_run_script(hdev, fileid):
    msg = [PFX_CMD_RUN_SCRIPT, fileid]
    return msg_transaction(hdev, msg)


def cmd_file_dir(hdev, req, params=None):
    msg = [PFX_CMD_FILE_DIR, req]
    if params is not None:
        msg.extend(params)
    return msg_transaction(hdev, msg)


def cmd_get_current_state(hdev):
    msg = [PFX_CMD_GET_CURRENT_STATE]
    return msg_transaction(hdev, msg)


def cmd_raw(hdev, msg):
    return msg_transaction(hdev, msg)


def i2c_write(brick, add, data):
    brick.send_raw_icd_command([PFX_CMD_WRITE_I2C, 0x30, add, data])


def i2c_read(brick, add):
    res = brick.send_raw_icd_command([PFX_CMD_READ_I2C, 0x30, add])
    nbytes = res[1]
    return res[2 : 2 + nbytes]


def flash_read(brick, add, num_bytes):
    rbytes = []
    for x in range(0, num_bytes, 60):
        msg = [PFX_CMD_READ_FLASH]
        msg.extend(uint32_to_bytes(add + x))
        msg.append(60)
        res = brick.send_raw_icd_command(msg)
        rbytes.extend(res[1:61])
    return rbytes[0:num_bytes]



#! /usr/bin/env python3
"""
pfxevents - print the contents of the event/action table
"""
import argparse

from rich import print
from rich.console import Console
from rich.table import Table

from pfxbrick import *
from pfxbrick.pfxdict import evtid_dict


def banner(title, raw=False):
    table = Table(show_header=False, width=76)
    table.add_row("[bold orange1]%s" % (title))
    print(table)
    if raw:
        print("[bold yellow]Add     Evt    Ch   Action")
    else:
        print("[bold yellow]Add   Evt               Ch    Action")


def main():
    parser = argparse.ArgumentParser(
        description="PFx Brick print event/action table",
        prefix_chars="-+",
    )
    parser.add_argument(
        "-cs",
        "--clear-speed",
        action="store_true",
        default=False,
        help="Clear actions for speed remote",
    )
    parser.add_argument(
        "-cj",
        "--clear-joystick",
        action="store_true",
        default=False,
        help="Clear actions for joystick remote",
    )
    parser.add_argument(
        "-cu",
        "--clear-startup",
        action="store_true",
        default=False,
        help="Clear startup actions",
    )
    parser.add_argument(
        "-ca",
        "--clear-all",
        action="store_true",
        default=False,
        help="Clear all actions",
    )
    parser.add_argument(
        "-r",
        "--raw",
        action="store_true",
        default=False,
        help="Show event/action table in raw numeric format",
    )
    parser.add_argument(
        "-i",
        "--ir",
        action="store_true",
        default=False,
        help="Show event/action table grouped by IR channel",
    )
    parser.add_argument(
        "-s",
        "--serialno",
        default=None,
        help="Specify PFx Brick with serial number (if more than one connected)",
    )
    args = parser.parse_args()
    argsd = vars(args)

    b = get_one_pfxbrick(argsd["serialno"])
    r = b.open()
    if not r:
        exit()
    b.refresh_file_dir()
    if argsd["clear_all"]:
        for a in range(0x4C):
            b.clear_action_by_address(a)
    elif argsd["clear_speed"]:
        for a in range(0x1C):
            b.clear_action_by_address(a)
    elif argsd["clear_joystick"]:
        for a in range(0x1C, 0x34):
            b.clear_action_by_address(a)
    elif argsd["clear_startup"]:
        for a in range(0x3C, 0x44):
            b.clear_action_by_address(a)

    al = 0
    ah = 0
    ac = 0
    for add in range(0x4C):
        if add == 0:
            banner("IR Speed Remote", argsd["raw"])
        elif add == 0x1C:
            banner("Dual Joystick Remote", argsd["raw"])
        elif add == 0x34:
            banner("EV3 Remote", argsd["raw"])
        elif add == 0x3C:
            banner("Startup Events", argsd["raw"])
        elif add == 0x44:
            banner("Other Events", argsd["raw"])
        evt, ch = address_to_evtch(add)
        if evt <= 6:
            if add > 0:
                if add % 7 == 0:
                    ah += 1
                    al = 0
                else:
                    al += 4
            ac = al + ah
        elif evt > 6 and evt <= 12:
            if add == 0x1C:
                al = 0
                ah = 28
            elif (add + 26) % 6 == 0:
                ah += 1
                al = 0
            else:
                al += 4
            ac = al + ah
        else:
            ac = add
        if argsd["ir"]:
            add = ac
        evt, ch = address_to_evtch(add)
        a = b.get_action_by_address(add)
        if argsd["raw"]:
            fmt = "0x%02X [blue]%2d[/blue] [blue](0x%02X)[/blue] [aquamarine3]%2d[/aquamarine3] : "
            ab = a.to_bytes()
            for i in range(16):
                if ab[i] == 0x00:
                    fmt = fmt + "[black]%02X "
                else:
                    fmt = fmt + "[white]%02X "
                if i == 7:
                    fmt = fmt + " "

            print(fmt % (add, add, evt, ch, *a.to_bytes()))
        else:
            if add < 0x3C:
                es = shorter_str(evtid_dict[evt])
            else:
                es = shorter_str(evtid_dict[add])
            if a.is_empty():
                print(
                    "0x%02X: [bold blue]%-16s[/bold blue] Ch [aquamarine3]%d[/aquamarine3] : ---"
                    % (add, es, ch + 1)
                )
            else:
                vs = shorter_str(a.verbose_line_str(b))
                print(
                    "0x%02X: [bold blue]%-16s[/bold blue] Ch [aquamarine3]%d[/aquamarine3] : %s"
                    % (add, es, ch + 1, vs)
                )

    b.close()


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
#
# Copyright (C) 2018  Fx Bricks Inc.
# This file is part of the pfxbrick python module.
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# PFx Brick internal state helpers

from pfxbrick import *
from pfxbrick.pfxhelpers import *


class PFxMotorState:
    def __init__(self):
        self.dir = "Forward"
        self.target_speed = 0
        self.current_speed = 0
        self.pwm_speed = 0

    def __str__(self):
        return "Dir: %7s Target speed: 0x%02X  Current: 0x%02X  PWM: 0x%02X" % (
            self.dir,
            self.target_speed,
            self.current_speed,
            self.pwm_speed,
        )

    def from_bytes(self, msg):
        self.dir = "Reverse" if (msg[0] & 0x01) else "Forward"
        self.target_speed = msg[1]
        self.current_speed = msg[2]
        self.pwm_speed = msg[3]


class PFxLightState:
    def __init__(self):
        self.active = False
        self.target_level = 0
        self.current_level = 0

    def __str__(self):
        return "Active: %5s Target level: 0x%02X  Current level: 0x%02X" % (
            self.active,
            self.target_level,
            self.current_level,
        )


class PFxAudioChannel:
    def __init__(self):
        self.mode = 0
        self.file_id = PFX_FILE_INVALID_ID

    def __str__(self):
        return "Mode: %d Current file: %3d" % (self.mode, self.file_id)


class PFxBTState:
    def __init__(self):
        self.sleep = 0
        self.state = 0
        self.flags = 0
        self.error = 0
        self.features = 0
        self.services = 0
        self.auth = 0
        self.tx_count = 0
        self.rx_count = 0
        self.msg_len = 0
        self.msg_prefix = []

    def __str__(self):
        s = []
        s.append(
            "Sleep: %d State: 0x%04X  Flags: 0x%04X  Err: 0x%04X"
            % (self.sleep, self.state, self.flags, self.error)
        )
        s.append(
            "BLE Feautures: 0x%04X  Services: 0x%04X  Auth: 0x%04X"
            % (self.features, self.services, self.auth)
        )
        s.append("Tx Count: %d  Rx Count: %d" % (self.tx_count, self.rx_count))
        mb = " ".join(["0x%02X" % (x) for x in self.msg_prefix])
        s.append("Last message len: %d <%s>" % (self.msg_len, mb))
        return "\n".join(s)

    def from_bytes(self, msg):
        msg = msg[1:]
        self.sleep = msg[1]
        self.state = uint16_toint(msg[2:4])
        self.flags = uint16_toint(msg[4:6])
        self.error = uint16_toint(msg[6:8])
        self.features = uint16_toint(msg[8:10])
        self.services = uint16_toint(msg[10:12])
        self.auth = uint16_toint(msg[12:14])
        self.tx_count = uint16_toint(msg[14:16])
        self.rx_count = uint16_toint(msg[16:18])
        self.msg_len = msg[19]
        self.msg_prefix = msg[20:28]


class PFxFSState:
    def __init__(self):
        self.file_count = 0
        self.open_files = 0
        self.task_state = 0
        self.flags = 0
        self.erase_sector = 0
        self.init_timemout = 0
        self.auto_sync_dir = 0
        self.auto_sync_map = 0
        self.sector_capacity = 0
        self.free_sectors = 0
        self.empty_sectors = 0

    def __str__(self):
        s = []
        s.append(
            "State: 0x%02X  Flags: 0x%02X  Files: %d  Open: %d"
            % (self.task_state, self.flags, self.file_count, self.open_files)
        )
        s.append(
            "Erase: 0x%04X  InitTime: 0x%04X  SyncDir: 0x%04X  SyncMap: 0x%04X"
            % (
                self.erase_sector,
                self.init_timemout,
                self.auto_sync_dir,
                self.auto_sync_map,
            )
        )
        s.append(
            "Sector Capacity: 0x%04X  Free: 0x%04X  Empty: 0x%04X"
            % (self.sector_capacity, self.free_sectors, self.empty_sectors)
        )
        return "\n".join(s)

    def from_bytes(self, msg):
        self.file_count = msg[1]
        self.open_files = uint16_toint(msg[18:20])
        self.task_state = msg[2]
        self.flags = msg[3]
        self.erase_sector = uint16_toint(msg[4:6])
        self.init_timemout = uint16_toint(msg[6:8])
        self.auto_sync_dir = uint16_toint(msg[8:10])
        self.auto_sync_map = uint16_toint(msg[10:12])
        self.sector_capacity = uint16_toint(msg[12:14])
        self.free_sectors = uint16_toint(msg[14:16])
        self.empty_sectors = uint16_toint(msg[16:18])


class PFxState:
    """
    PFx Brick internal data state container class.

    This class is used to store internal state data obtained with the `PFX_CMD_GET_CURRENT_STATE`
    ICD command.  This internal state data can be useful for monitoring, debugging, and
    for building apps which can use this data for enhanced feedback of the PFx Brick runtime
    state and behaviour.

    Attributes:
        brightness (:obj:`int`): current global brightness value

        volume (:obj:`int`): current audio volume level

        millisec_count (:obj:`int`): PFx Brick millisecond counter value

        slow_count (:obj:`int`): PFx Brick 1 second counter value

        status_latch1 (:obj:`int`): PFx Brick status latch 1 bit flag state

        status_latch2 (:obj:`int`): PFx Brick status latch 2 bit flag state

        audio_peak (:obj:`int`): current audio peak level based on current playback

        audio_notch (:obj:`int`): current motor notch value for indexed audio playback

        fs_state (:obj:`int`): PFx Brick file system state flags

        script_state (:obj:`int`): PFx Brick script engine state

        script_line (:obj:`int`): current running script line pointer

        motors ([:obj:`PFxMotorState`]): motor channel runtime state container class

        lights ([:obj:`PFxLightState`]): light channel runtime state container class

        audio_ch ([:obj:`PFxAudioChannel`]): audio channel runtime state container class

    """

    def __init__(self):
        self.brightness = 0
        self.volume = 0
        self.motors = [PFxMotorState() for ch in range(PFX_MOTOR_CHANNELS)]
        self.lights = [PFxLightState() for ch in range(PFX_LIGHT_CHANNELS)]
        self.audio_ch = [PFxAudioChannel() for ch in range(PFX_AUDIO_CHANNELS)]
        self.lightmask = 0
        self.millisec_count = 0
        self.slow_count = 0
        self.status_latch1 = 0
        self.status_latch2 = 0
        self.audio_peak = 0
        self.audio_notch = 0
        self.fs_state = 0
        self.script_state = 0
        self.script_line = 0
        self.motor_ptr = 0
        self.motor_pwm_ptr = 0
        self.motor_rate_ptr = 0
        self.trig_change_dir_state = 0
        self.trig_set_off_state = 0
        self.trig_rapid_accel_state = 0
        self.trig_rapid_decel_state = 0
        self.trig_brake_state = 0
        self.filesys = PFxFSState()
        self.bt = PFxBTState()

    def from_bytes(self, msg):
        """
        Converts the message string bytes read from the PFx Brick into
        the corresponding data members of this class.
        """
        self.brightness = msg[1]
        self.volume = msg[2]
        for ch, idx in zip(self.motors, [3, 7]):
            ch.from_bytes(msg[idx : idx + 4])
        self.motor_ptr = msg[11]
        self.motor_pwm_ptr = msg[12]
        self.motor_rate_ptr = msg[13]
        self.trig_change_dir_state = msg[14]
        self.trig_set_off_state = msg[15]
        self.trig_rapid_accel_state = msg[16]
        self.trig_rapid_decel_state = msg[17]
        self.trig_brake_state = msg[18]
        self.lightmask = msg[19]
        for i, ch in enumerate(self.lights):
            ch.target_level = msg[21 + i]
            ch.current_level = msg[33 + i]
            if i > 7:
                mask = 1 << (i - 8)
                ch.active = True if (msg[20] & mask) else False
            else:
                mask = 1 << i
                ch.active = True if (msg[19] & mask) else False
        for ch, idx in zip(self.audio_ch, [45, 47, 49, 51]):
            ch.mode = msg[idx]
            ch.file_id = msg[idx + 1]
        self.millisec_count = uint16_toint(msg[53:55])
        self.slow_count = uint16_toint(msg[55:57])
        self.status_latch1 = msg[57]
        self.status_latch2 = msg[58]
        self.fs_state = msg[59]
        self.audio_peak = msg[60]
        self.audio_notch = msg[61]
        self.script_state = msg[62]
        self.script_line = msg[63]

    def __str__(self):
        s = []
        s.append("Brightness: %3d  Volume: %3d" % (self.brightness, self.volume))
        s.append(
            "Slow count: %4X  Millisec count: %4X"
            % (self.slow_count, self.millisec_count)
        )
        s.append(
            "File system: %2X Script exec: %2X Script line: %3d"
            % (self.fs_state, self.script_state, self.script_line)
        )
        s.append(
            "Status 1: %2X  Status 2: %2X" % (self.status_latch1, self.status_latch2)
        )
        for i, ch in enumerate(self.motors):
            s.append("Motor Ch %d : %s" % (i + 1, str(ch)))
        for i, ch in enumerate(self.lights):
            s.append("Light Ch %2d : %s" % (i + 1, str(ch)))
        for i, ch in enumerate(self.audio_ch):
            s.append("Audio Ch %d : %s" % (i + 1, str(ch)))
        s.append(
            "Audio Peak: %3d  Audio notch: %d" % (self.audio_peak, self.audio_notch)
        )
        return "\n".join(s)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Learn more: https://github.com/fx-bricks/pfx-brick-py/setup.py

import os
import platform
import sys

import setuptools

PACKAGE_NAME = "pfxbrick"
MINIMUM_PYTHON_VERSION = "3.6"


def check_python_version():
    """Exit when the Python version is too low."""
    if platform.python_version() < MINIMUM_PYTHON_VERSION:
        sv = "Python %s+ is required, not %s" % (MINIMUM_PYTHON_VERSION, sys.version)
        sys.exit(sv)


def read_package_variable(key, filename="__init__.py"):
    """Read the value of a variable from the package without importing."""
    module_path = os.path.join(PACKAGE_NAME, filename)
    with open(module_path) as module:
        for line in module:
            parts = line.strip().split(" ", 2)
            if parts[:-1] == [key, "="]:
                return parts[-1].strip("'")
    sys.exit("'{0}' not found in '{1}'".format(key, module_path))


def build_description():
    """Build a description for the project from documentation files."""
    try:
        readme = open("README.rst", encoding="utf8").read()
        changelog = open("CHANGELOG.rst", encoding="utf8").read()
    except IOError:
        return "<placeholder>"
    else:
        return readme + "\n" + changelog


setuptools.setup(
    name=read_package_variable("__project__"),
    version=read_package_variable("__version__"),
    description="A python API for host control of USB connected PFx Bricks.",
    url="https://github.com/fx-bricks/pfx-brick-py",
    author="Fx Bricks Inc.",
    author_email="info@fxbricks.com",
    packages=setuptools.find_packages(),
    long_description=build_description(),
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=["hidapi", "bleak>=0.12", "rich"],
    entry_points={
        "console_scripts": [
            "pfxcat=pfxbrick.scripts.pfxcat:main",
            "pfxdir=pfxbrick.scripts.pfxdir:main",
            "pfxget=pfxbrick.scripts.pfxget:main",
            "pfxput=pfxbrick.scripts.pfxput:main",
            "pfxrm=pfxbrick.scripts.pfxrm:main",
            "pfxrename=pfxbrick.scripts.pfxrename:main",
            "pfxdump=pfxbrick.scripts.pfxdump:main",
            "pfxmonitor=pfxbrick.scripts.pfxmonitor:main",
            "pfxtest=pfxbrick.scripts.pfxtest:main",
            "pfxinfo=pfxbrick.scripts.pfxinfo:main",
            "pfxplay=pfxbrick.scripts.pfxplay:main",
            "pfxrun=pfxbrick.scripts.pfxrun:main",
            "pfxfat=pfxbrick.scripts.pfxfat:main",
            "pfxrestart=pfxbrick.scripts.pfxrestart:main",
            "pfxevents=pfxbrick.scripts.pfxevents:main",
            "pfxscan=pfxbrick.scripts.pfxscan:main",
        ],
    },
)



#! /usr/bin/env python3
"""
pfxinfo - get identity and status information from PFx Brick
"""
import argparse
from datetime import datetime

from rich import print
from rich.console import Console
from rich.table import Table

from pfxbrick import *

console = Console()


def main():
    parser = argparse.ArgumentParser(
        description="Show information for all attached PFx Bricks"
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store_true",
        default=False,
        help="Show configuration details",
    )
    args = parser.parse_args()
    argsd = vars(args)

    bricks = find_bricks()
    for brick in bricks:
        b = PFxBrick(brick)
        b.open()
        b.get_status()
        icd = b.get_icd_rev()
        if b.status == 0x00:
            name = b.get_name()
        else:
            name = "Service Mode"

        table = Table(show_header=True, header_style="bold blue")
        bid = "[light_slate_blue]%s [bold cyan]%s" % (b.product_id, b.product_desc)
        table.add_column(bid, width=72)
        table.add_row("Serial Number         : [bold cyan]%s" % (b.serial_no))
        table.add_row("ICD Version           : [bold green]%s" % (icd))
        table.add_row(
            "Firmware Version      : [bold green]%s [reset]build [bold green]%s"
            % (b.firmware_ver, b.firmware_build)
        )
        table.add_row("USB vendor ID         : [cyan]0x%04X" % (b.usb_vid))
        table.add_row("USB product ID        : [cyan]0x%04X" % (b.usb_pid))
        table.add_row(
            "Status                : 0x%02X %s" % (b.status, get_status_str(b.status))
        )
        if b.error:
            table.add_row(
                "Errors                : [red]0x%02X %s"
                % (b.error, get_error_str(b.error))
            )
        else:
            table.add_row(
                "Errors                : [green]0x%02X %s"
                % (b.error, get_error_str(b.error))
            )

        table.add_row("Name                  : [bold yellow]%s" % (name))

        console.print(table)
        if b.status == 0x00:
            b.get_current_state()
            b.get_fs_state()

            if argsd["config"]:
                b.get_config()
                console.print(str(b.config))


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
"""
pfxdump - show the raw contents of the PFx Brick flash memory
"""
import argparse

from pfxbrick import *


def main():
    parser = argparse.ArgumentParser(
        description="PFx Brick dump flash memory contents",
        prefix_chars="-+",
    )
    parser.add_argument(
        "address",
        metavar="address",
        type=str,
        help="base address to start showing contents",
    )
    parser.add_argument(
        "bytes",
        metavar="bytes",
        type=int,
        help="number of bytes to show",
        default=256,
    )
    parser.add_argument(
        "-s",
        "--serialno",
        default=None,
        help="Specify PFx Brick with serial number (if more than one connected)",
    )
    args = parser.parse_args()
    argsd = vars(args)

    b = get_one_pfxbrick(argsd["serialno"])
    b.open()
    r = b.open()
    if not r:
        exit()
    rb = flash_read(b, int(argsd["address"], 16), int(argsd["bytes"]))
    pprint_bytes(rb, str(argsd["address"]))
    b.close()


if __name__ == "__main__":
    main()



"""Package for PFx Brick

   isort:skip_file
"""
import os

# fmt: off
__project__ = 'pfxbrick'
__version__ = '0.8.6'

# API version

VERSION = __project__ + "-" + __version__

# ICD version this API is compliant with
#   for more detail, consult:
#   https://www.fxbricks.com/downloads/PFxBrickICD-Rev3.39.pdf
#   and
#   https://github.com/fx-bricks/pfx-brick-dev

ICD_REV = '3.39'

script_dir = os.path.dirname(__file__)

from .pfx import *
from .pfxaction import PFxAction
from .pfxconfig import PFxConfig
from .pfxstate import PFxState
from .pfxfiles import (PFxDir, PFxFile, fs_copy_file_from, fs_copy_file_to,
                       fs_get_fileid_from_name, fs_remove_file)
from .pfxbrick import PFxBrick, find_bricks
from .pfxble import PFxBrickBLE, ble_device_scanner, find_ble_pfxbricks
from .pfxdict import *
from .pfxexceptions import *
from .pfxmsg import *
from .pfxhelpers import *

# fmt: on
